from __future__ import annotations

from prettyqt import network
from prettyqt.qt import QtCore
from prettyqt.utils import datatypes


class HttpPart(network.QHttpPart):
    """Holds a body part to be used inside a HTTP multipart MIME message."""

    def set_body(self, body: datatypes.ByteArrayType):
        body = datatypes.to_bytearray(body)
        self.setBody(body)

    def set_headers(self, headers: dict[str, str]):
        for k, v in headers.items():
            self.setRawHeader(
                QtCore.QByteArray(k.encode()), QtCore.QByteArray(v.encode())
            )

    def set_header(self, name: str, value: network.networkrequest.KnownHeaderStr):
        self.setHeader(network.networkrequest.KNOWN_HEADER[name], value)


if __name__ == "__main__":
    part = HttpPart()



from __future__ import annotations

from prettyqt.qt import QtQml


class QmlParserStatusMixin:
    pass


class QmlParserStatus(QmlParserStatusMixin, QtQml.QQmlParserStatus):
    """Updates on the QML parser state."""



from __future__ import annotations

from prettyqt import constants, core, itemmodels, widgets


class FuzzyCompleter(widgets.Completer):
    def __init__(self, parent):
        super().__init__(parent)
        parent.setEditable(True)
        parent.set_insert_policy("no_insert")
        parent.installEventFilter(self)
        self.set_completion_mode("popup")
        self._local_completion_prefix = ""
        self._source_model = None
        self._filter_proxy = itemmodels.FuzzyFilterProxyModel(self)
        self._filter_proxy.set_match_color(None)
        self._using_original_model = False

    def setModel(self, model):
        self._source_model = model
        self._filter_proxy = itemmodels.FuzzyFilterProxyModel(self)
        self._filter_proxy.set_match_color(None)
        self._filter_proxy.setSourceModel(self._source_model)
        super().setModel(self._filter_proxy)
        self._using_original_model = True

    def eventFilter(self, source: widgets.QLineEdit, event: core.QEvent) -> bool:
        match event.type():
            case core.QEvent.Type.FocusIn:
                source.clearEditText()
            case core.QEvent.Type.KeyPress:
                key = event.key()
                if key == constants.Key.Key_Enter:
                    text = source.currentText()
                    source.setCompleter(None)
                    source.setEditText(text)
                    source.setCompleter(source.comp)
        return super().eventFilter(source, event)

    def updateModel(self):
        if not self._using_original_model:
            self._filter_proxy.setSourceModel(self._source_model)

        self._filter_proxy.set_search_term(self._local_completion_prefix)

    def splitPath(self, path):
        self._local_completion_prefix = path
        self.updateModel()
        if self._filter_proxy.rowCount() == 0:
            self._using_original_model = False
            model = core.StringListModel([path])
            self._filter_proxy.setSourceModel(model)
            return [path]

        return []


if __name__ == "__main__":
    from prettyqt import custom_widgets

    app = widgets.app()
    widget = custom_widgets.StandardIconsWidget()
    widget.show()
    app.exec()



from __future__ import annotations

from collections.abc import Sequence
import inspect
import logging
from typing import ClassVar

from prettyqt import constants, core, gui, itemmodels


logger = logging.getLogger(__name__)

SOURCE_FONT = gui.Font.mono(as_qt=True)


class FrameInfoModel(itemmodels.ListMixin, core.AbstractTableModel):
    """Model to display a list of inspect.Frameinfos / inspect.Tracebacks."""

    HEADER: ClassVar = [
        "Filename",
        "Line number",
        "Function",
        "Code context",
        "Index",
        "Positions",
    ]
    SUPPORTS = Sequence[inspect.FrameInfo | inspect.Traceback]

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (inspect.FrameInfo(), *_) | (inspect.Traceback(), *_):
                return True
            case _:
                return False

    def columnCount(self, parent=None) -> int:
        return len(self.HEADER)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE:
                return self.HEADER[section]
            case constants.VERTICAL, constants.DISPLAY_ROLE:
                return str(section)
        return None

    def data(  # noqa: PLR0911
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        field = self.items[index.row()]
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                return field.filename
            case constants.DISPLAY_ROLE, 1:
                return field.lineno
            case constants.DISPLAY_ROLE, 2:
                return field.function
            case constants.DISPLAY_ROLE, 3:
                lines = []
                pretty = ""
                for line in field.code_context:
                    while line.startswith(" "):
                        pretty += "·"
                        line = line[1:]
                    pretty += line
                    lines.append(pretty)
                return "\n".join(lines)
            case constants.FONT_ROLE, 3:
                return SOURCE_FONT
            case constants.DISPLAY_ROLE, 4:
                return field.index
            case constants.DISPLAY_ROLE, 5:
                p = field.positions
                return f"{p.lineno} - {p.end_lineno}, {p.col_offset} {p.end_col_offset}"
            case constants.ALIGNMENT_ROLE, _:
                return constants.ALIGN_CENTER_LEFT


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TableView()
    view.set_word_wrap(True)
    view.set_icon("mdi.folder")
    with app.debug_mode():
        stack = inspect.stack()
        model = FrameInfoModel(items=stack)
        view.set_model(model)
        view.set_selection_behavior("rows")
        view.set_edit_triggers("all")
        view.set_delegate("editor", column=1)
        view.show()
        view.resize(500, 300)
        app.exec()



"""Webenginewidgets module.

contains QtWebEngineWidgets-based classes
"""

import logging


logger = logging.getLogger(__name__)

try:
    from prettyqt.qt.QtWebEngineWidgets import *  # noqa: F403

    from .webengineview import WebEngineView
    from prettyqt.qt import QtWebEngineWidgets

    QT_MODULE = QtWebEngineWidgets
    __all__ = [
        "WebEngineView",
    ]
except ModuleNotFoundError:
    logger.warning("Tried to import WebEngine module but not installed.")



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal

from prettyqt import core, gui, iconprovider, widgets
from prettyqt.utils import bidict, datatypes


if TYPE_CHECKING:
    from collections.abc import Iterable


mod = widgets.QComboBox

InsertPolicyStr = Literal[
    "no_insert",
    "top",
    "current",
    "bottom",
    "after_current",
    "before_current",
    "alphabetically",
]

INSERT_POLICY: bidict[InsertPolicyStr, mod.InsertPolicy] = bidict(
    no_insert=mod.InsertPolicy.NoInsert,
    top=mod.InsertPolicy.InsertAtTop,
    current=mod.InsertPolicy.InsertAtCurrent,
    bottom=mod.InsertPolicy.InsertAtBottom,
    after_current=mod.InsertPolicy.InsertAfterCurrent,
    before_current=mod.InsertPolicy.InsertBeforeCurrent,
    alphabetically=mod.InsertPolicy.InsertAlphabetically,
)

SizeAdjustPolicyStr = Literal[
    "content",
    "first_show",
    # "min_length",
    "min_length_with_icon",
]

SIZE_ADJUST_POLICY: bidict[SizeAdjustPolicyStr, mod.SizeAdjustPolicy] = bidict(
    content=mod.SizeAdjustPolicy.AdjustToContents,
    first_show=mod.SizeAdjustPolicy.AdjustToContentsOnFirstShow,
    # min_length=mod.SizeAdjustPolicy.AdjustToMinimumContentsLength,
    min_length_with_icon=mod.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon,
)


class ComboBoxMixin(widgets.WidgetMixin):
    """Combobox that lets the user select a font family."""

    value_changed = core.Signal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.currentIndexChanged.connect(self._index_changed)

    def __len__(self) -> int:
        return self.count()

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "insertPolicy": INSERT_POLICY,
            "sizeAdjustPolicy": SIZE_ADJUST_POLICY,
        }
        return maps

    # def __reduce__(self):
    #     return type(self), (), self.__getstate__()

    def _index_changed(self, index: int):
        # data = self.itemData(index)
        data = self.get_value()
        self.value_changed.emit(data)

    def add_items(
        self, items: Iterable[str | tuple | list] | Mapping[Any, str], default=...
    ):
        if isinstance(items, Mapping):
            for k, v in items.items():
                self.addItem(v, userData=k)
        else:
            for i in items:
                if isinstance(i, tuple | list):
                    self.add(*i)
                else:
                    self.addItem(i, i)
        if default is not ...:
            self.set_value(default)

    def add(self, label: str, data=..., icon: datatypes.IconType = None):
        if data is ...:
            data = label
        if icon is not None:
            icon = iconprovider.get_icon(icon)
            self.addItem(gui.Icon(icon), label, userData=data)
        else:
            self.addItem(label, userData=data)

    def get_item_icon(self, index: int) -> gui.Icon | None:
        icon = self.itemIcon(index)
        return None if icon.isNull() else gui.Icon(icon)

    def set_editable(self, editable: bool):
        self.setEditable(editable)
        if self.completer() is None and editable:
            self.setCompleter(widgets.Completer(self))

    def set_insert_policy(self, policy: InsertPolicyStr | mod.InsertPolicy):
        """Set insert policy.

        Args:
            policy: insert policy to use
        """
        self.setInsertPolicy(INSERT_POLICY.get_enum_value(policy))

    def get_insert_policy(self) -> InsertPolicyStr:
        """Return insert policy.

        Returns:
            insert policy
        """
        return INSERT_POLICY.inverse[self.insertPolicy()]

    def set_size_adjust_policy(self, policy: SizeAdjustPolicyStr | mod.SizeAdjustPolicy):
        """Set size adjust policy.

        Args:
            policy: size adjust policy to use
        """
        self.setSizeAdjustPolicy(SIZE_ADJUST_POLICY.get_enum_value(policy))

    def get_size_adjust_policy(self) -> SizeAdjustPolicyStr:
        """Return size adjust policy.

        Returns:
            size adjust policy
        """
        return SIZE_ADJUST_POLICY.inverse[self.sizeAdjustPolicy()]

    def set_icon_size(self, size: datatypes.SizeType):
        """Set size of the icons."""
        self.setIconSize(datatypes.to_size(size))

    def get_icon_size(self) -> core.Size:
        return core.Size(self.iconSize())

    def set_min_char_length(self, chars: int):
        self.setMinimumContentsLength(chars)

    def get_value(self) -> Any:
        # if all(self.itemData(i) is None for i in range(self.count())):
        #     return self.currentText()
        # else:
        #     return self.currentData()
        return self.currentData()

    def set_value(self, value: Any):
        self.set_data(value)

    def set_text(self, text: str):
        self.setCurrentText(text)

    def set_data(self, data: Any):
        idx = self.findData(data)
        if idx == -1:
            msg = "invalid data"
            raise ValueError(msg)
        self.setCurrentIndex(idx)

    def text(self) -> str:
        return self.currentText()

    def hide_completer(self):
        """Hides the completion widget."""
        completer = widgets.Completer(self)
        self.setCompleter(completer)


class ComboBox(ComboBoxMixin, widgets.QComboBox):
    """Combined button and popup list."""


if __name__ == "__main__":
    app = widgets.app()
    widget = ComboBox()
    w = ComboBox()
    widget.add("test", data="aa", icon="mdi.timer")
    widget.add("test2", data="aa2", icon="mdi.timer")
    widget.show()
    app.exec()



from __future__ import annotations

import contextlib
from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict


ExitStatusStr = Literal["normal", "crash"]

EXIT_STATUS: bidict[ExitStatusStr, core.QProcess.ExitStatus] = bidict(
    normal=core.QProcess.ExitStatus.NormalExit,
    crash=core.QProcess.ExitStatus.CrashExit,
)

InputChannelModeStr = Literal["managed", "forwarded"]

INPUT_CHANNEL_MODES: bidict[InputChannelModeStr, core.QProcess.InputChannelMode] = bidict(
    managed=core.QProcess.InputChannelMode.ManagedInputChannel,
    forwarded=core.QProcess.InputChannelMode.ForwardedInputChannel,
)

ProcessChannelStr = Literal["standard", "error"]

PROCESS_CHANNELS: bidict[ProcessChannelStr, core.QProcess.ProcessChannel] = bidict(
    standard=core.QProcess.ProcessChannel.StandardOutput,
    error=core.QProcess.ProcessChannel.StandardError,
)

ProcessChannelModeStr = Literal[
    "separate", "merged", "forwarded", "forwarded_error", "forwarded_output"
]

PROCESS_CHANNEL_MODES: bidict[ProcessChannelModeStr, core.QProcess.ProcessChannelMode] = (
    bidict(
        separate=core.QProcess.ProcessChannelMode.SeparateChannels,
        merged=core.QProcess.ProcessChannelMode.MergedChannels,
        forwarded=core.QProcess.ProcessChannelMode.ForwardedChannels,
        forwarded_error=core.QProcess.ProcessChannelMode.ForwardedErrorChannel,
        forwarded_output=core.QProcess.ProcessChannelMode.ForwardedOutputChannel,
    )
)

ProcessErrorStr = Literal[
    "failed_to_start", "crashed", "timed_out", "write", "read_error", "unknown_error"
]

PROCESS_ERRORS: bidict[ProcessErrorStr, core.QProcess.ProcessError] = bidict(
    failed_to_start=core.QProcess.ProcessError.FailedToStart,
    crashed=core.QProcess.ProcessError.Crashed,
    timed_out=core.QProcess.ProcessError.Timedout,
    write=core.QProcess.ProcessError.WriteError,
    read_error=core.QProcess.ProcessError.ReadError,
    unknown_error=core.QProcess.ProcessError.UnknownError,
)


ProcessStateStr = Literal["not_running", "starting", "running"]

PROCESS_STATES: bidict[ProcessStateStr, core.QProcess.ProcessState] = bidict(
    not_running=core.QProcess.ProcessState.NotRunning,
    starting=core.QProcess.ProcessState.Starting,
    running=core.QProcess.ProcessState.Running,
)


class Process(core.IODeviceMixin, core.QProcess):
    """Used to start external programs and to communicate with them."""

    def set_read_channel(
        self, channel: ProcessChannelStr | core.QProcess.ProcessChannelMode
    ):
        """Set the input channel channel.

        Args:
            channel: channel to set
        """
        self.setReadChannel(PROCESS_CHANNELS.get_enum_value(channel))

    def get_read_channel(self) -> ProcessChannelStr:
        return PROCESS_CHANNELS.inverse[self.readChannel()]

    def close_read_channel(self, channel: ProcessChannelStr):
        self.closeReadChannel(PROCESS_CHANNELS[channel])

    def set_input_channel_mode(
        self, mode: InputChannelModeStr | core.QProcess.InputChannelMode
    ):
        """Set the input channel mode.

        Args:
            mode: mode to set
        """
        self.setInputChannelMode(INPUT_CHANNEL_MODES.get_enum_value(mode))

    def get_input_channel_mode(self) -> InputChannelModeStr:
        return INPUT_CHANNEL_MODES.inverse[self.inputChannelMode()]

    def set_process_channel_mode(
        self, mode: ProcessChannelModeStr | core.QProcess.ProcessChannelMode
    ):
        """Set the process channel mode.

        Args:
            mode: mode to set
        """
        self.setProcessChannelMode(PROCESS_CHANNEL_MODES.get_enum_value(mode))

    def get_process_channel_mode(self) -> ProcessChannelModeStr:
        return PROCESS_CHANNEL_MODES.inverse[self.processChannelMode()]

    def set_state(self, state: ProcessStateStr | core.QProcess.ProcessState):
        """Set the process state.

        Args:
            state: state to set
        """
        self.setProcessState(PROCESS_STATES.get_enum_value(state))

    def get_state(self) -> ProcessStateStr:
        return PROCESS_STATES.inverse[self.state()]

    def get_error(self) -> ProcessErrorStr:
        return PROCESS_ERRORS.inverse[self.error()]

    def get_exit_status(self) -> ExitStatusStr:
        return EXIT_STATUS.inverse[self.exitStatus()]

    def get_process_environment(self) -> core.ProcessEnvironment:
        return core.ProcessEnvironment(self.processEnvironment())

    @contextlib.contextmanager
    def edit_process_environment(self):
        env = self.get_process_environment()
        yield env
        self.setProcessEnvironment(env)


if __name__ == "__main__":
    process = core.Process()
    env = process.get_process_environment()
    print(env.items())



from __future__ import annotations

from typing import Literal, Self

from prettyqt import core, gui
from prettyqt.utils import bidict, datatypes, get_repr, serializemixin


MODE = bidict(
    normal=gui.QIcon.Mode.Normal,
    disabled=gui.QIcon.Mode.Disabled,
    active=gui.QIcon.Mode.Active,
    selected=gui.QIcon.Mode.Selected,
)

ModeStr = Literal["normal", "disabled", "active", "selected"]

STATE = bidict(off=gui.QIcon.State.Off, on=gui.QIcon.State.On)

StateStr = Literal["off", "on"]


class Icon(serializemixin.SerializeMixin, gui.QIcon):
    """Scalable icons in different modes and states."""

    def __repr__(self):
        return get_repr(self)

    def __bool__(self):
        return not self.isNull()

    def __getstate__(self):
        pixmap = self.pixmap(256, 256)
        return bytes(gui.Pixmap(pixmap))

    def __setstate__(self, ba):
        px = gui.Pixmap()
        px.__setstate__(ba)
        self.add_pixmap(px)

    @classmethod
    def for_color(cls, color_str: str) -> Self:
        color = gui.Color(color_str)
        bitmap = gui.Pixmap(16, 16)
        bitmap.fill(color)
        return cls(bitmap)

    @classmethod
    def from_char(cls, char: str, background="black", color="white") -> Self:
        """Create a QIcon with a given character."""
        icon = cls()
        for size in (16, 32, 64):
            px = gui.Pixmap.create_char(
                char, background=background, color=color, size=size
            )
            icon.addPixmap(px)
        return icon

    @classmethod
    def from_image(cls, image: gui.QImage) -> Self:
        return cls(gui.Pixmap.fromImage(image))

    def get_available_sizes(
        self,
        mode: ModeStr | gui.QIcon.Mode = "normal",
        state: StateStr | gui.QIcon.State = "off",
    ) -> list[core.Size]:
        m = MODE.get_enum_value(mode)
        s = STATE.get_enum_value(state)
        return [core.Size(i) for i in self.availableSizes(m, s)]

    def add_pixmap(
        self,
        data: core.QByteArray | gui.QPixmap | bytes,
        mode: ModeStr | gui.QIcon.Mode = "normal",
        state: StateStr | gui.QIcon.State = "off",
    ):
        if isinstance(data, bytes):
            data = core.QByteArray(data)
        if isinstance(data, core.QByteArray):
            pixmap = gui.QPixmap()
            pixmap.loadFromData(data)
        else:
            pixmap = data
        self.addPixmap(pixmap, MODE.get_enum_value(mode), STATE.get_enum_value(state))

    def get_pixmap(
        self,
        size: datatypes.SizeType | int,
        mode: ModeStr | gui.QIcon.Mode = "normal",
        state: StateStr | gui.QIcon.State = "off",
    ) -> gui.QPixmap:
        sz = datatypes.to_size(size)
        return self.pixmap(sz, MODE.get_enum_value(mode), STATE.get_enum_value(state))

    def get_actual_size(
        self,
        size: datatypes.SizeType,
        mode: ModeStr | gui.QIcon.Mode = "normal",
        state: StateStr | gui.QIcon.State = "off",
    ) -> core.Size:
        sz = datatypes.to_size(size)
        m = MODE.get_enum_value(mode)
        s = STATE.get_enum_value(state)
        actual_size = self.actualSize(sz, m, s)
        return core.Size(actual_size)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    icon = Icon.from_char("A")
    window = widgets.MainWindow()
    window.set_icon(icon)
    window.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core, qthelp


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class HelpEngineCoreMixin(core.ObjectMixin):
    """The core functionality of the help system."""

    def get_file_data(self, url: core.QUrl) -> bytes:
        return self.fileData(url).data()

    def get_files(
        self,
        namespace_name: str,
        filter_name: str,
        extension_filter: str | None = None,
    ) -> list[core.Url]:
        if extension_filter is None:
            extension_filter = ""
        return [
            core.Url(i) for i in self.files(namespace_name, filter_name, extension_filter)
        ]

    def find_file(self, url: datatypes.UrlType) -> core.Url:
        if not isinstance(url, core.QUrl):
            url = core.QUrl(url)
        return core.Url(self.findFile(url))

    def get_filter_engine(self) -> qthelp.HelpFilterEngine:
        return qthelp.HelpFilterEngine(self.filterEngine())


class HelpEngineCore(HelpEngineCoreMixin, qthelp.QHelpEngineCore):
    pass


if __name__ == "__main__":
    engine = HelpEngineCore("")
    engine.get_files("a", "b")



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import constants, itemmodels, widgets


if TYPE_CHECKING:
    from prettyqt import core


class SubsequenceCompleter(widgets.Completer):
    """QCompleter specialised for subsequence matching."""

    def __init__(self, *args):
        super().__init__(*args)
        self.completion_search_term = ""
        self.source_model = None
        self.set_completion_mode("unfiltered_popup")
        self.proxy_model = itemmodels.SubsequenceSortFilterProxyModel(parent=self)
        self.proxy_model.setFilterCaseSensitivity(self.caseSensitivity())
        self.set_case_sensitive(False)
        self._force_next_update = True
        self.path_updated.connect(self._on_path_updated)
        self.setCompletionRole(constants.DISPLAY_ROLE)

    def setModel(self, model: core.QAbstractItemModel):
        self.source_model = model
        self.proxy_model = itemmodels.SubsequenceSortFilterProxyModel(parent=self)
        self.proxy_model.setFilterCaseSensitivity(self.caseSensitivity())
        self.proxy_model.set_search_term(self.completion_search_term)
        self.proxy_model.setSourceModel(self.source_model)
        super().setModel(self.proxy_model)
        self.proxy_model.invalidateRowsFilter()
        # self.proxy_model.sort(0)
        self._force_next_update = True

    # def __getattr__(self, key):
    #     return getattr(self.proxy_model, key)

    def set_case_sensitive(self, value: bool):
        super().set_case_sensitive(value)
        self.proxy_model.set_filter_case_sensitive(value)

    def _on_path_updated(self, path: str):
        if path == self.completion_search_term:
            return
        self.completion_search_term = path
        count = self.completionCount()
        if count or len(self.completion_search_term) <= 1 or self._force_next_update:
            self.proxy_model.set_search_term(self.completion_search_term)
        if count > 1:
            self.proxy_model.sort(0)
        self._force_next_update = False
        self.proxy_model.invalidateRowsFilter()  # force sorting/filtering


if __name__ == "__main__":
    from prettyqt import widgets
    from prettyqt.itemmodels.importlibdistributionmodel import (
        ImportlibDistributionModel,
    )

    app = widgets.app()
    source_model = ImportlibDistributionModel.from_package("prettyqt")
    completer = SubsequenceCompleter()
    completer.setModel(source_model)

    lineedit = widgets.LineEdit()
    lineedit.set_completer(completer)
    widget = widgets.Widget()
    widget.set_layout("vertical")
    widget.box.add(lineedit)
    table = widgets.TableView()
    widget.box.add(table)
    table.set_model(source_model)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt import constants, widgets
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from prettyqt import core


area = widgets.QAbstractScrollArea

SizeAdjustPolicyStr = Literal["content", "first_show", "ignored"]

SIZE_ADJUST_POLICY: bidict[SizeAdjustPolicyStr, area.SizeAdjustPolicy] = bidict(
    content=area.SizeAdjustPolicy.AdjustToContents,
    first_show=area.SizeAdjustPolicy.AdjustToContentsOnFirstShow,
    ignored=area.SizeAdjustPolicy.AdjustIgnored,
)


class AbstractScrollAreaMixin(widgets.FrameMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setHorizontalScrollBar(widgets.ScrollBar(parent=self))
        self.setVerticalScrollBar(widgets.ScrollBar(parent=self))

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "horizontalScrollBarPolicy": constants.SCROLLBAR_POLICY,
            "sizeAdjustPolicy": SIZE_ADJUST_POLICY,
            "verticalScrollBarPolicy": constants.SCROLLBAR_POLICY,
        }
        return maps

    @property
    def h_scrollbar(self) -> widgets.ScrollBar:
        return self.horizontalScrollBar()

    @h_scrollbar.setter
    def h_scrollbar(self, scrollbar: widgets.ScrollBar):
        self.setHorizontalScrollBar(scrollbar)

    @property
    def v_scrollbar(self) -> widgets.ScrollBar:
        return self.verticalScrollBar()

    @v_scrollbar.setter
    def v_scrollbar(self, scrollbar: widgets.ScrollBar):
        self.setVerticalScrollBar(scrollbar)

    def scroll_by_pixels(self, x: int = 0, y: int = 0):
        new_x = self.h_scrollbar.value() + x
        x_val = max(min(new_x, self.h_scrollbar.maximum()), self.h_scrollbar.minimum())
        new_y = self.h_scrollbar.value() + y
        y_val = max(min(new_y, self.v_scrollbar.maximum()), self.v_scrollbar.minimum())
        self.h_scrollbar.setValue(x_val)
        self.v_scrollbar.setValue(y_val)

    def set_size_adjust_policy(self, policy: SizeAdjustPolicyStr | area.SizeAdjustPolicy):
        """Set size adjust policy.

        Args:
            policy: size adjust policy to use
        """
        self.setSizeAdjustPolicy(SIZE_ADJUST_POLICY.get_enum_value(policy))

    def get_size_adjust_policy(self) -> SizeAdjustPolicyStr:
        """Return size adjust policy.

        Returns:
            size adjust policy
        """
        return SIZE_ADJUST_POLICY.inverse[self.sizeAdjustPolicy()]

    def set_scrollbar_smooth(
        self,
        value: bool = True,
        animation_duration: int = 500,
        easing: core.easingcurve.TypeStr | core.QEasingCurve.Type = "out_cubic",
    ):
        from prettyqt import custom_widgets

        if value:
            self.h_scrollbar = custom_widgets.SmoothScrollBar(
                "horizontal",
                parent=self,
                animation_duration=animation_duration,
                easing=easing,
            )
            self.v_scrollbar = custom_widgets.SmoothScrollBar(
                "vertical",
                parent=self,
                animation_duration=animation_duration,
                easing=easing,
            )
        else:
            self.h_scrollbar = widgets.ScrollBar(parent=self)
            self.v_scrollbar = widgets.ScrollBar(parent=self)

    def set_scrollbar_policy(
        self, mode: constants.ScrollBarPolicyStr | constants.ScrollBarPolicy
    ):
        """Set the policy for both scrollbars.

        Args:
            mode: visibilty to set
        """
        self.setHorizontalScrollBarPolicy(constants.SCROLLBAR_POLICY.get_enum_value(mode))
        self.setVerticalScrollBarPolicy(constants.SCROLLBAR_POLICY.get_enum_value(mode))

    def set_horizontal_scrollbar_policy(
        self, mode: constants.ScrollBarPolicyStr | constants.ScrollBarPolicy
    ):
        """Set the horizontal scrollbar visibility.

        Args:
            mode: visibilty to set
        """
        self.setHorizontalScrollBarPolicy(constants.SCROLLBAR_POLICY[mode])

    def get_horizontal_scrollbar_policy(self) -> constants.ScrollBarPolicyStr:
        return constants.SCROLLBAR_POLICY.inverse[self.horizontalScrollBarPolicy()]

    def set_vertical_scrollbar_policy(
        self, mode: constants.ScrollBarPolicyStr | constants.ScrollBarPolicy
    ):
        """Set the vertical scrollbar visibility.

        Args:
            mode: visibilty to set
        """
        self.setVerticalScrollBarPolicy(constants.SCROLLBAR_POLICY.get_enum_value(mode))

    def get_vertical_scrollbar_policy(self) -> constants.ScrollBarPolicyStr:
        return constants.SCROLLBAR_POLICY.inverse[self.verticalScrollBarPolicy()]

    def set_scrollbar_width(self, width: int):
        """Set the width for both scrollbars.

        Args:
            width: width in pixels
        """
        self.set_horizontal_scrollbar_width(width)
        self.set_vertical_scrollbar_width(width)

    def set_horizontal_scrollbar_width(self, width: int):
        """Set the horizontal scrollbar width.

        Args:
            width: width in pixels
        """
        with self.h_scrollbar.edit_stylesheet() as ss:
            ss.QScrollBar.horizontal.height.setValue(f"{width}px")

    def set_vertical_scrollbar_width(self, width: int):
        """Set the vertical scrollbar width.

        Args:
            width: width in pixels
        """
        with self.v_scrollbar.edit_stylesheet() as ss:
            ss.QScrollBar.horizontal.height.setValue(f"{width}px")

    def scroll_to_top(self):
        """Scroll to the top of the scroll area."""
        self.verticalScrollBar().scroll_to_min()

    def scroll_to_bottom(self):
        """Scroll to the bottom of the scroll area."""
        self.verticalScrollBar().scroll_to_max()

    def set_viewport_margins(self, margins: int):
        self.setViewportMargins(margins, margins, margins, margins)

    def add_scrollbar_widget(
        self,
        widget: widgets.QWidget,
        alignment: constants.AlignmentStr | constants.AlignmentFlag,
    ):
        alignment = constants.ALIGNMENTS.get_enum_value(alignment)
        self.addScrollBarWidget(widget, alignment)


class AbstractScrollArea(AbstractScrollAreaMixin, widgets.QAbstractScrollArea):
    """Scrolling area with on-demand scroll bars."""


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.PlainTextEdit()
    widget.show()
    app.exec()



from __future__ import annotations

import logging
import os
from typing import Literal

from qtconsole.rich_jupyter_widget import RichJupyterWidget

from prettyqt import core, gui, widgets


CompletionmodeStr = Literal["droplist", "plain", "ncurses"]
PagingStr = Literal["inside", "hsplit", "vsplit", "custom", "none"]
KindStr = Literal["plain", "rich"]

# disables 'Please pass -Xfrozen_modules=off' warning
os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"
os.environ["JUPYTER_PLATFORM_DIRS"] = "1"

logger = logging.getLogger(__name__)


class BaseIPythonWidget(RichJupyterWidget, widgets.WidgetMixin):
    """Convenience class for a live IPython console widget running out-of-process."""

    code_executed = core.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(
            gui_completion="droplist",  # 'plain', 'droplist', 'ncurses'
            kind="rich",  # 'plain', 'rich', only applies when no custom control set.
            paging="vsplit",  # h  'inside', 'hsplit', 'vsplit', 'custom', 'none'
            # custom_control=widgets.TextEdit,
            # custom_page_control=widgets.TextEdit,
        )
        self.banner = "IPython Console"
        self.font_size = 6
        # self.gui_completion = "droplist"
        widgets.Application.call_on_exit(self.shutdown)
        # self.exit_requested.connect(self.shutdown)

        widgets.Application.styleHints().colorSchemeChanged.connect(
            self.adjust_style_to_palette
        )
        self.adjust_style_to_palette()

    @classmethod
    def setup_example(cls):
        return None

    def set_completion_mode(self, mode: CompletionmodeStr):
        self._set_completion_widget(mode)

    def get_completion_mode(self) -> CompletionmodeStr:
        return self.gui_completion  # type: ignore

    @property
    def texteditor(self) -> widgets.QPlainTextEdit:  # to shut up linters
        return self._control  # type: ignore

    def adjust_style_to_palette(self):
        """Adjust coloring of the terminal to current palette."""
        pal = widgets.Application.get_palette()
        style = "linux" if pal.is_dark() else "lightbg"
        self.set_default_style(style)

    def shutdown(self):
        """Stop IPython and cleanup."""
        return NotImplemented

    def buffer(self) -> str:
        """Get current code block."""
        return self.input_buffer

    def set_buffer(self, code: str) -> None:
        """Set code string to Jupyter QtConsole buffer."""
        self.input_buffer = ""
        if not isinstance(code, str):
            msg = f"Cannot set {type(code)}."
            raise ValueError(msg)  # noqa: TRY004
        cursor = self.texteditor.textCursor()
        lines = code.split("\n")
        for line in lines[:-1]:
            cursor.insertText(line + "\n")
            self._insert_continuation_prompt(cursor)  # insert "...:"
        cursor.insertText(lines[-1])

    def insert_text(self, text: str) -> None:
        cursor = self.texteditor.textCursor()
        cursor.insertText(text)

    def set_temp_text(self, text: str) -> None:
        cursor = self.texteditor.textCursor()
        cursor.removeSelectedText()
        pos = cursor.position()
        cursor.insertText(text)
        cursor.setPosition(pos)
        cursor.setPosition(pos + len(text), gui.TextCursor.MoveMode.KeepAnchor)
        self.texteditor.setTextCursor(cursor)

    def get_selected_text(self) -> str:
        """Return the selected text."""
        cursor = self.texteditor.textCursor()
        return cursor.selection().toPlainText()

    def clear(self):
        """Clear the terminal."""
        self.texteditor.clear()

    def print_text(self, text: str, before_prompt: bool = False):
        """Print some plain text to the console."""
        self._append_plain_text(text)

    def execute(
        self,
        source: str | None = None,
        hidden: bool = False,
        interactive: bool = False,
    ):
        """Execute a command in the frame of the console widget."""
        if source is None:
            source = self.input_buffer
        super().execute(source=source, hidden=hidden, interactive=interactive)
        self.code_executed.emit(source)

    completion_mode = core.Property(
        str,
        get_completion_mode,
        set_completion_mode,
        doc="Completion mode for the widget",
    )


if __name__ == "__main__":
    from prettyqt.custom_widgets import commandpalette

    app = widgets.app()
    cp = commandpalette.CommandPalette()
    app.set_style("fusion")
    console_widget = BaseIPythonWidget(app)
    console_widget.print_text("hallo")
    console_widget.show()
    console_widget.add_shortcut("Ctrl+P", cp.show)
    cp.populate_from_widget(console_widget)
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from bidict import bidict

from prettyqt import constants, core, itemmodels


if TYPE_CHECKING:
    from collections.abc import Callable, Mapping


class SliceMapRoleProxyModel(itemmodels.SliceIdentityProxyModel):
    """Proxy model for mapping one role to another.

    Mapping can be changed by passing a dictionary with source role as key and target_role
    as value.
    Py passing an optional converter function, values which are mapped can be modified.

    ### Example

    ```py
    source_model = FsSpecTreemodel("file")
    table = widgets.TableView()
    mapping = {source_model.Roles.ProtocolPathRole: constants.DISPLAY_ROLE}
    model = SliceMapRoleProxyModel(mapping, indexer=0, parent=table)
    model.setSourceModel(source_model)
    table.set_model(model)
    table.show()
    # or
    table.proxifier.map_role(source_model.Roles.ProtocolPathRole, constants.DISPLAY_ROLE)
    ```
    """

    ID = "map_role"
    ICON = "mdi.directions-fork"

    def __init__(
        self,
        mapping: Mapping[constants.ItemDataRole, constants.ItemDataRole],
        converter: Callable | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._mapping = bidict(mapping)
        self._converter = converter

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if role in self._mapping.inverse and self.indexer_contains(index):
            value = super().data(index, self._mapping.inverse[role])
            return self._converter(value) if self._converter else value
        return super().data(index, role)

    def set_mapping(
        self,
        mapping: Mapping[constants.ItemDataRole, constants.ItemDataRole],
    ):
        with self.reset_model():
            self._mapping = bidict(mapping)

    def get_mapping(self) -> Mapping[constants.ItemDataRole, constants.ItemDataRole]:
        return self._mapping

    mapping = core.Property(
        dict,
        get_mapping,
        set_mapping,
        doc="Mapping of ItemRoles",
    )


if __name__ == "__main__":
    from prettyqt import widgets
    from prettyqt.itemmodels import fsspecmodel

    source_model = fsspecmodel.FSSpecTreeModel("file")
    app = widgets.app()
    table = widgets.TableView()
    mapping = {
        source_model.Roles.ProtocolPathRole: constants.DISPLAY_ROLE,
    }
    model = SliceMapRoleProxyModel(mapping, indexer=0, parent=table)
    model.setSourceModel(source_model)
    table.set_model(model)
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from prettyqt import core, gui
from prettyqt.utils import get_repr


class TextObjectMixin(core.ObjectMixin):
    def __repr__(self):
        return get_repr(self)

    def get_format(self) -> gui.TextFormat:
        return gui.TextFormat(self.format())


class TextObject(TextObjectMixin, gui.QTextObject):
    pass


if __name__ == "__main__":
    doc = gui.TextDocument()
    obj = TextObject(doc)



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core


if TYPE_CHECKING:
    from prettyqt import constants, widgets


class Animator(core.Object):
    """Helper class to animate TabWidget / StackedWidget transistions and other.

    Required interface:

       * `currentIndex` / `setCurrentIndex`
       * `widget`
       * `indexOf`
       * `count`
       * `rect`
    """

    def __init__(self, widget):
        super().__init__(widget)
        self._widget = widget
        self.direction = "horizontal"
        self.speed = 500
        self.animation_type = "out_cubic"
        self.now = 0
        self.next = 0
        self.wrap = False
        self.pnow = core.Point(0, 0)
        self.active = False

    def set_direction(self, direction: constants.OrientationStr):
        self.direction = direction

    def set_speed(self, speed: int):
        self.speed = speed

    def set_animation(self, animation_type: core.easingcurve.TypeStr):
        self.animation_type = animation_type

    def set_wrap(self, wrap: bool):
        self.wrap = wrap

    @core.Slot()
    def slide_in_previous(self):
        now = self._widget.currentIndex()
        if self.wrap or now > 0:
            self.slide_in(now - 1)

    @core.Slot()
    def slide_in_next(self):
        now = self._widget.currentIndex()
        if self.wrap or now < (self._widget.count() - 1):
            self.slide_in(now + 1)

    def slide_in(self, idx: int | widgets.QWidget):
        if self.active:
            return
        if isinstance(idx, int):
            if idx > (self._widget.count() - 1):
                idx = idx % self._widget.count()
            elif idx < 0:
                idx = (idx + self._widget.count()) % self._widget.count()
            idx = self._widget.widget(idx)

        self.active = True

        _now = self._widget.currentIndex()
        _next = self._widget.indexOf(idx)

        if _now == _next:
            self.active = False
            return
        if hasattr(self._widget, "frameRect"):  # TabWidget doesnt have frameRect
            frame_rect = self._widget.frameRect()
        else:
            frame_rect = self._widget.rect()
        offsetx, offsety = (frame_rect.width(), frame_rect.height())
        self._widget.widget(_next).setGeometry(frame_rect)

        if self.direction == "horizontal":
            if _now < _next:
                offsetx, offsety = -offsetx, 0
            else:
                offsety = 0

        elif _now < _next:
            offsetx, offsety = 0, -offsety
        else:
            offsetx = 0
        pnext = self._widget.widget(_next).pos()
        pnow = self._widget.widget(_now).pos()
        self.pnow = pnow

        offset = core.Point(offsetx, offsety)
        self._widget.widget(_next).move(pnext - offset)
        self._widget.widget(_next).show()
        self._widget.widget(_next).raise_()

        anim_group = core.ParallelAnimationGroup(
            self._widget, finished=self._on_animation_done
        )

        for index, start, end in zip(
            (_now, _next), (pnow, pnext - offset), (pnow + offset, pnext)
        ):
            animation = core.PropertyAnimation(
                self._widget.widget(index),
                b"pos",
                duration=self.speed,
                easing_curve=self.animation_type,
                start_value=start,
                end_value=end,
            )
            anim_group.addAnimation(animation)

        self.next = _next
        self.now = _now
        self.active = True
        anim_group.start_animation(policy="delete")

    @core.Slot()
    def _on_animation_done(self):
        self._widget.setCurrentIndex(self.next)
        self._widget.widget(self.now).hide()
        self._widget.widget(self.now).move(self.pnow)
        self.active = False

    def fade_in(self, widget: int | widgets.QWidget):
        from prettyqt import custom_widgets

        widget = self._widget.widget(widget) if isinstance(widget, int) else widget
        self._widget.fader_widget = custom_widgets.FaderWidget(
            self._widget.currentWidget(), widget, self.speed
        )
        self._widget.setCurrentWidget(widget)



from __future__ import annotations

from prettyqt import core, gui, iconprovider, widgets
from prettyqt.utils import datatypes


class AbstractButtonMixin(widgets.WidgetMixin):
    """Abstract button widget base class, providing functionality common to buttons."""

    def __bool__(self):
        return self.isChecked()

    def set_icon(self, icon: datatypes.IconType):
        """Set the icon for the button.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon)
        self.setIcon(icon)

    def get_icon(self) -> gui.Icon | None:
        icon = self.icon()
        return None if icon.isNull() else gui.Icon(icon)

    def set_style_icon(
        self,
        icon: widgets.style.StandardPixmapStr | widgets.QStyle.StandardPixmap,
        size: datatypes.SizeType = 15,
    ):
        """Set theme icon for the button.

        Args:
            icon: icon to use
            size: icon size
        """
        qicon = self.style().standardIcon(
            widgets.style.STANDARD_PIXMAP.get_enum_value(icon), None, self
        )
        self.set_icon(qicon)
        self.setIconSize(datatypes.to_size(size))

    def set_shortcut(self, shortcut: datatypes.KeySequenceType):
        self.setShortcut(datatypes.to_keysequence(shortcut))

    def get_shortcut(self) -> gui.KeySequence:
        return gui.KeySequence(
            self.shortcut().toString(), gui.KeySequence.SequenceFormat.PortableText
        )

    def setText(self, text: str):
        if not self.objectName() and widgets.app().is_debug():
            self.setObjectName(text)
        super().setText(text)

    set_text = setText

    def set_icon_size(self, size: datatypes.SizeType):
        """Set size of the icon."""
        self.setIconSize(datatypes.to_size(size))

    def get_icon_size(self) -> core.Size:
        return core.Size(self.iconSize())

    def get_value(self) -> bool:
        return self.isChecked()

    def set_value(self, value: bool):
        self.setChecked(value)

    # def _color(self):
    #     return self.palette().color(QtGui.QPalette.ColorRole.ButtonText)

    # def _set_color(self, qcolor: QtGui.QColor):
    #     palette = self.palette()
    #     palette.setColor(QtGui.QPalette.ColorRole.ButtonText, qcolor)
    #     self.setPalette(palette)

    # color = core.Property(QtGui.QColor, _color, _set_color)

    # @core.Property(QtGui.QColor)
    # def background_color(self):
    #     return self.palette().color(QtGui.QPalette.ColorRole.Button)

    # @background_color.setter
    # def background_color(self, qcolor: QtGui.QColor):
    #     palette = self.palette()
    #     palette.setColor(QtGui.QPalette.ColorRole.Button, qcolor)
    #     self.setPalette(palette)


class AbstractButton(AbstractButtonMixin, widgets.QAbstractButton):
    pass


if __name__ == "__main__":
    app = widgets.app()
    widget = AbstractButton()
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt.qt import QtCore
from prettyqt.utils import get_repr


class JsonValue(QtCore.QJsonValue):
    """Encapsulates a value in JSON."""

    def __str__(self):
        return str(self.toVariant())

    def __repr__(self):
        return get_repr(self, self.toVariant())


if __name__ == "__main__":
    from prettyqt import core

    doc = core.JsonDocument.from_variant(dict(a="b"))
    new = core.JsonDocument()
    new.setObject(doc.object())
    val = doc["a"]



from __future__ import annotations

from prettyqt import core, gui, widgets


class BoolDictToolButton(widgets.ToolButton):
    value_changed = core.Signal(dict)

    def __init__(
        self,
        *args,
        dct: dict[str, str] | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.button_menu = widgets.Menu(triggered=self._on_menu_click)
        self.setMenu(self.button_menu)
        self.set_popup_mode("instant")
        if dct:
            self.set_dict(dct)

    def _on_menu_click(self):
        self.value_changed.emit(self.as_dict())

    def __getitem__(self, key: str) -> bool:  # type: ignore
        return self.button_menu[key].isChecked()

    def __setitem__(self, key: str, value: bool):
        self.button_menu[key].setChecked(value)
        self.value_changed.emit(self.as_dict())

    def set_dict(self, dct: dict[str, str]):
        self.button_menu.clear()
        for k, v in dct.items():
            action = gui.Action(text=v, checkable=True, object_name=k)
            self.button_menu.add(action)
        self.value_changed.emit(self.as_dict())

    def as_dict(self) -> dict[str, bool]:
        return {act.objectName(): act.isChecked() for act in self.button_menu}


if __name__ == "__main__":
    app = widgets.app()
    dct = dict(a="test", b="test2")
    w = BoolDictToolButton(text="Title", dct=dct)
    w.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import gui, iconprovider, widgets


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class UndoView(widgets.ListViewMixin, widgets.QUndoView):
    """Displays the contents of a QUndoStack."""

    def __getitem__(self, index: int) -> gui.QUndoCommand:
        return self.stack().command(index)

    def set_clean_icon(self, icon: datatypes.IconType):
        """Set the icon for the clean button.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon)
        self.setCleanIcon(icon)

    def set_value(self, value: gui.QUndoGroup | gui.QUndoStack):
        if isinstance(value, gui.QUndoGroup):
            self.setGroup(value)
        else:
            self.setStack(value)


if __name__ == "__main__":
    app = widgets.app()
    view = UndoView()
    stack = gui.UndoStack()
    stack.add_command("test", redo=lambda: print("redo"), undo=lambda: print("undo"))
    view.setStack(stack)
    view.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, gui, widgets


class IconEdit(widgets.Widget):
    value_changed = core.Signal(gui.Icon)

    def __init__(self, *args, object_name: str = "icon_edit", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)

    def _on_value_change(self):
        self._value = self.get_value()
        self.value_changed.emit(self._value)

    def get_value(self) -> gui.Icon:
        return self._value

    def set_value(self, value: gui.Icon):
        self._value = value

    value = core.Property(
        gui.Icon,
        get_value,
        set_value,
        user=True,
        doc="Currently chosen icon",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = IconEdit(window_title="Test")
    widget.set_value(gui.Icon())
    widget.value_changed.connect(print)
    widget.show()
    app.exec()



from __future__ import annotations

import contextlib
import functools
import weakref

from prettyqt import core, widgets
from prettyqt.qt import QtGui


class FlashEffect(widgets.GraphicsColorizeEffect):
    def __init__(self, duration: int = 1000, color="green", **kwargs):
        super().__init__(**kwargs)
        self._flash_animation = core.PropertyAnimation(self, b"color")
        self._flash_animation.setStartValue(QtGui.QColor(0, 0, 0, 0))
        self._flash_animation.setEndValue(QtGui.QColor(0, 0, 0, 0))
        self._flash_animation.setLoopCount(1)

        def remove_flash_animation(widget_ref: weakref.ref[widgets.QWidget]):
            """Remove flash animation from widget.

            Parameters
            ----------
            widget_ref : QWidget
                Any Qt widget.
            """
            if widget_ref() is None:
                return
            widget = widget_ref()
            with contextlib.suppress(RuntimeError):
                widget.setGraphicsEffect(None)

        # let's make sure to remove the animation from the widget because
        # if we don't, the widget will actually be black and white.
        self._flash_animation.finished.connect(
            functools.partial(remove_flash_animation, weakref.ref(self.parent()))
        )

    def flash(self, duration: int = 1000, color="red"):
        """Add flash animation to widget to highlight certain action.

        Parameters
        ----------
        widget : QWidget
            Any Qt widget.
        duration : int
            Duration of the flash animation.
        color : Array
            Color of the flash animation. By default, we use red.
        """
        color = QtGui.QColor(color)
        # now  set an actual time for the flashing and an intermediate color
        self._flash_animation.setDuration(duration)
        self._flash_animation.setKeyValueAt(0.1, color)
        self._flash_animation.start()


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    btn = widgets.PushButton("ts")
    effect = FlashEffect(parent=btn)
    btn.show()
    app.sleep(2)
    btn.setGraphicsEffect(effect)
    effect.flash()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Self

from prettyqt import constants, widgets
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from prettyqt.qt import QtGui
    from prettyqt.utils import datatypes


PopupModeStr = Literal["delayed", "menu_button", "instant"]

POPUP_MODE: bidict[PopupModeStr, widgets.QToolButton.ToolButtonPopupMode] = bidict(
    delayed=widgets.QToolButton.ToolButtonPopupMode.DelayedPopup,
    menu_button=widgets.QToolButton.ToolButtonPopupMode.MenuButtonPopup,
    instant=widgets.QToolButton.ToolButtonPopupMode.InstantPopup,
)


class ToolButton(widgets.AbstractButtonMixin, widgets.QToolButton):
    """Quick-access button to commands or options, usually used inside a QToolBar."""

    def __getitem__(self, item: str) -> QtGui.QAction:
        menu = self.menu()
        return menu[item]

    # def set_menu(self, menu: widgets.QMenu):
    #     menu.setParent(self)
    #     self.setMenu(menu)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "toolButtonStyle": constants.TOOLBUTTON_STYLE,
            "arrowType": constants.ARROW_TYPE,
            "popupMode": POPUP_MODE,
        }
        return maps

    @classmethod
    def for_menu(cls, menu: widgets.QMenu, icon: datatypes.IconType = None) -> Self:
        btn = cls()
        btn.setMenu(menu)
        # btn.set_title(menu.title())
        btn.set_popup_mode("instant")
        btn.set_icon(icon)
        return btn

    def set_default_action(self, action: QtGui.QAction):
        self.setDefaultAction(action)

    def set_popup_mode(
        self, mode: PopupModeStr | widgets.QToolButton.ToolButtonPopupMode
    ):
        """Set the popup mode of the toolbutton.

        Args:
            mode: popup mode to use
        """
        self.setPopupMode(POPUP_MODE.get_enum_value(mode))

    def get_popup_mode(self) -> PopupModeStr:
        """Return popup mode.

        Returns:
            popup mode
        """
        return POPUP_MODE.inverse[self.popupMode()]

    def set_arrow_type(self, mode: constants.ArrowTypeStr | constants.ArrowType):
        """Set the arrow type of the toolbutton.

        Args:
            mode: arrow type to use
        """
        self.setArrowType(constants.ARROW_TYPE.get_enum_value(mode))

    def get_arrow_type(self) -> constants.ArrowTypeStr:
        """Return arrow type.

        Returns:
            arrow type
        """
        return constants.ARROW_TYPE.inverse[self.arrowType()]

    def set_style(self, style: constants.ToolButtonStyleStr | constants.ToolButtonStyle):
        """Set the toolbutton style.

        Args:
            style: style to use
        """
        self.setToolButtonStyle(constants.TOOLBUTTON_STYLE.get_enum_value(style))

    def get_style(self) -> constants.ToolButtonStyleStr:
        """Return toolbutton style.

        Returns:
            toolbutton style
        """
        return constants.TOOLBUTTON_STYLE.inverse[self.toolButtonStyle()]


if __name__ == "__main__":
    app = widgets.app()
    w = ToolButton()
    # icon = gui.Icon()
    # icon.add_pixmap("mdi.timer")
    # icon.add_pixmap("mdi.folder", state="on")
    w.set_icon("mdi.timer", "mdi.folder")
    w.show()
    app.exec()



from __future__ import annotations

from ctypes import POINTER, Structure, byref, c_bool, c_int, cdll, pointer, sizeof
from ctypes.wintypes import BOOL, DWORD, HRGN, LONG, LPCVOID, ULONG
import enum
import sys
from typing import ClassVar, SupportsInt

import win32con
import win32gui


class WINDOWCOMPOSITIONATTRIB(enum.Enum):
    WCA_ACCENT_POLICY = 19
    WCA_USEDARKMODECOLORS = 26


class ACCENT_STATE(enum.Enum):  # noqa: N801
    ACCENT_DISABLED = 0
    ACCENT_ENABLE_ACRYLICBLURBEHIND = 4  # Acrylic effect
    ACCENT_ENABLE_HOSTBACKDROP = 5  # Mica effect


class DWMWINDOWATTRIBUTE(enum.Enum):
    DWMWA_NCRENDERING_POLICY = 2


class DWMNCRENDERINGPOLICY(enum.Enum):
    DWMNCRP_DISABLED = 1


class ACCENT_POLICY(Structure):  # noqa: N801
    _fields_: ClassVar = [
        ("AccentState", DWORD),
        ("AccentFlags", DWORD),
        ("GradientColor", DWORD),
        ("AnimationId", DWORD),
    ]


class WINDOWCOMPOSITIONATTRIBDATA(Structure):
    _fields_: ClassVar = [
        ("Attribute", DWORD),
        ("Data", POINTER(ACCENT_POLICY)),
        ("SizeOfData", ULONG),
    ]


class MARGINS(Structure):
    _fields_: ClassVar = [
        ("cxLeftWidth", c_int),
        ("cxRightWidth", c_int),
        ("cyTopHeight", c_int),
        ("cyBottomHeight", c_int),
    ]


class DWM_BLURBEHIND(Structure):  # noqa: N801
    _fields_: ClassVar = [
        ("dwFlags", DWORD),
        ("fEnable", BOOL),
        ("hRgnBlur", HRGN),
        ("fTransitionOnMaximized", BOOL),
    ]


class WindowsEffects:
    """Class for applying Windows effects."""

    def __init__(self):
        # Initialize structure
        self.accent_policy = ACCENT_POLICY()
        self.win_comp_attr_data = WINDOWCOMPOSITIONATTRIBDATA()
        self.win_comp_attr_data.Attribute = (
            WINDOWCOMPOSITIONATTRIB.WCA_ACCENT_POLICY.value
        )
        self.win_comp_attr_data.SizeOfData = sizeof(self.accent_policy)
        self.win_comp_attr_data.Data = pointer(self.accent_policy)

        # Declare the function signature of the API
        user_32 = cdll.LoadLibrary("user32")
        self.set_win_comp_attr = user_32.SetWindowCompositionAttribute
        dwm_api = cdll.LoadLibrary("dwmapi")
        self.dwm_ext_frame_into_client_area = dwm_api.DwmExtendFrameIntoClientArea
        self.dwm_enable_blur_behind_win = dwm_api.DwmEnableBlurBehindWindow
        self.dwm_set_win_attr = dwm_api.DwmSetWindowAttribute

        self.set_win_comp_attr.argtypes = [c_int, POINTER(WINDOWCOMPOSITIONATTRIBDATA)]
        self.dwm_ext_frame_into_client_area.argtypes = [c_int, POINTER(MARGINS)]
        self.dwm_enable_blur_behind_win.argtypes = [c_int, POINTER(DWM_BLURBEHIND)]
        self.dwm_set_win_attr.argtypes = [c_int, DWORD, LPCVOID, DWORD]

        self.set_win_comp_attr.restype = c_bool
        self.dwm_ext_frame_into_client_area.restype = LONG
        self.dwm_enable_blur_behind_win.restype = LONG
        self.dwm_set_win_attr.restype = LONG

    def add_acrylic_effect(
        self,
        h_wnd: SupportsInt,
        gradient_color: str,
        enable_shadow: bool = True,
        animation_id: int = 0,
    ):
        gradient_color = "".join(gradient_color[i : i + 2] for i in range(6, -1, -2))
        gradient_color = DWORD(int(gradient_color, base=16))
        accent_flags = DWORD(0x20 | 0x40 | 0x80 | 0x100) if enable_shadow else DWORD(0)
        animation_id = DWORD(animation_id)

        self.accent_policy.AccentState = (
            ACCENT_STATE.ACCENT_ENABLE_ACRYLICBLURBEHIND.value
        )
        self.accent_policy.AccentFlags = accent_flags
        self.accent_policy.GradientColor = gradient_color
        self.accent_policy.AnimationId = animation_id
        self.set_win_comp_attr(int(h_wnd), pointer(self.win_comp_attr_data))

    def add_mica_effect(self, h_wnd: SupportsInt, dark_mode: bool = False):
        h_wnd = int(h_wnd)
        self.accent_policy.AccentState = ACCENT_STATE.ACCENT_ENABLE_HOSTBACKDROP.value
        self.set_win_comp_attr(h_wnd, pointer(self.win_comp_attr_data))

        if dark_mode:
            self.win_comp_attr_data.Attribute = (
                WINDOWCOMPOSITIONATTRIB.WCA_USEDARKMODECOLORS.value
            )
            self.set_win_comp_attr(h_wnd, pointer(self.win_comp_attr_data))
            self.win_comp_attr_data.Attribute = (
                WINDOWCOMPOSITIONATTRIB.WCA_ACCENT_POLICY.value
            )

        if sys.getwindowsversion().build >= 22523:  # noqa: PLR2004
            self.dwm_set_win_attr(h_wnd, 38, byref(c_int(2)), 4)
        else:
            self.dwm_set_win_attr(h_wnd, 1029, byref(c_int(1)), 4)

    def remove_background_effect(self, h_wnd: SupportsInt):
        self.accent_policy.AccentState = ACCENT_STATE.ACCENT_DISABLED.value
        self.set_win_comp_attr(int(h_wnd), pointer(self.win_comp_attr_data))

    def add_shadow_effect(self, h_wnd: SupportsInt):
        margins = MARGINS(-1, -1, -1, -1)
        self.dwm_ext_frame_into_client_area(int(h_wnd), byref(margins))

    def remove_shadow_effect(self, h_wnd: SupportsInt):
        self.dwm_set_win_attr(
            int(h_wnd),
            DWMWINDOWATTRIBUTE.DWMWA_NCRENDERING_POLICY.value,
            byref(c_int(DWMNCRENDERINGPOLICY.DWMNCRP_DISABLED.value)),
            4,
        )

    @staticmethod
    def add_window_animation(h_wnd: SupportsInt):
        h_wnd = int(h_wnd)
        style = win32gui.GetWindowLong(h_wnd, win32con.GWL_STYLE)
        win32gui.SetWindowLong(
            h_wnd,
            win32con.GWL_STYLE,
            style
            | win32con.WS_MINIMIZEBOX
            | win32con.WS_MAXIMIZEBOX
            | win32con.WS_CAPTION
            | win32con.WS_THICKFRAME,
            # | win32con.CS_DBLCLKS
            # | win32con.WS_POPUP
            # | win32con.WS_SYSMENU
        )

    def add_blur_behind_window(self, h_wnd: SupportsInt):
        blur_behind = DWM_BLURBEHIND(1, True, 0, False)
        self.dwm_enable_blur_behind_win(int(h_wnd), byref(blur_behind))



from __future__ import annotations

from ctypes import Structure, byref, c_int, sizeof, windll
from ctypes.wintypes import DWORD, HWND, LPARAM, RECT, UINT
import os
import sys
from typing import ClassVar, SupportsInt

import win32api
from win32comext.shell import shellcon
import win32con
import win32gui


class APPBARDATA(Structure):
    _fields_: ClassVar = [
        ("cbSize", DWORD),
        ("hWnd", HWND),
        ("uCallbackMessage", UINT),
        ("uEdge", UINT),
        ("rc", RECT),
        ("lParam", LPARAM),
    ]


def raise_to_top(h_wnd: SupportsInt):
    h_wnd = int(h_wnd)
    # set to always-on-top and disable it again. that way windows stays in front
    flag = win32con.SWP_NOMOVE | win32con.SWP_NOSIZE | win32con.SWP_SHOWWINDOW
    win32gui.SetWindowPos(h_wnd, win32con.HWND_TOPMOST, 0, 0, 0, 0, flag)
    win32gui.SetWindowPos(h_wnd, win32con.HWND_NOTOPMOST, 0, 0, 0, 0, flag)


def set_window_nonresizable(h_wnd: SupportsInt):
    h_wnd = int(h_wnd)
    style = win32gui.GetWindowLong(h_wnd, win32con.GWL_STYLE)
    style &= ~win32con.WS_SIZEBOX
    style &= ~win32con.WS_THICKFRAME
    style &= ~win32con.WS_MAXIMIZEBOX
    win32gui.SetWindowLong(h_wnd, win32con.GWL_STYLE, style)


def is_window_resizable(h_wnd: SupportsInt) -> bool:
    h_wnd = int(h_wnd)
    style = win32api.GetWindowLong(h_wnd, win32con.GWL_STYLE)
    return style & win32con.WS_SIZEBOX != 0


def is_maximized(h_wnd: SupportsInt) -> bool:
    h_wnd = int(h_wnd)
    window_placement = win32gui.GetWindowPlacement(h_wnd)
    return window_placement[1] == win32con.SW_MAXIMIZE if window_placement else False


def start_system_move(h_wnd: SupportsInt) -> None:
    win32gui.ReleaseCapture()
    win32api.SendMessage(
        int(h_wnd), win32con.WM_SYSCOMMAND, win32con.SC_MOVE | win32con.HTCAPTION, 0
    )  # pyside6-frameless-window uses "+" instead of "|"?


def get_monitor_info(h_wnd, dw_flags):
    if monitor := win32api.MonitorFromWindow(h_wnd, dw_flags):
        return win32api.GetMonitorInfo(monitor)
    return None


def is_full_screen(h_wnd):
    if not h_wnd:
        return False
    h_wnd = int(h_wnd)

    win_rect = win32gui.GetWindowRect(h_wnd)
    if not win_rect:
        return False

    monitor_info = get_monitor_info(h_wnd, win32con.MONITOR_DEFAULTTOPRIMARY)
    if not monitor_info:
        return False

    monitor_rect = monitor_info["Monitor"]
    return all(i == j for i, j in zip(win_rect, monitor_rect))


def get_resize_border_thickness():
    result = win32api.GetSystemMetrics(
        win32con.SM_CXSIZEFRAME
    ) + win32api.GetSystemMetrics(92)

    if result > 0:
        return result

    b_result = c_int(0)
    windll.dwmapi.DwmIsCompositionEnabled(byref(b_result))
    return 8 if bool(b_result.value) else 4


class Taskbar:
    LEFT = 0
    TOP = 1
    RIGHT = 2
    BOTTOM = 3
    NO_POSITION = 4

    AUTO_HIDE_THICKNESS = 2

    @staticmethod
    def is_auto_hide():
        appbar_data = APPBARDATA(sizeof(APPBARDATA), 0, 0, 0, RECT(0, 0, 0, 0), 0)
        taskbar_state = windll.shell32.SHAppBarMessage(
            shellcon.ABM_GETSTATE, byref(appbar_data)
        )
        return taskbar_state == shellcon.ABS_AUTOHIDE

    @classmethod
    def get_position(cls, h_wnd):
        monitor_info = get_monitor_info(h_wnd, win32con.MONITOR_DEFAULTTONEAREST)
        if not monitor_info:
            return cls.NO_POSITION

        monitor = RECT(*monitor_info["Monitor"])
        appbar_data = APPBARDATA(sizeof(APPBARDATA), 0, 0, 0, monitor, 0)
        for position in (cls.LEFT, cls.TOP, cls.RIGHT, cls.BOTTOM):
            appbar_data.uEdge = position
            if windll.shell32.SHAppBarMessage(11, byref(appbar_data)):
                return position

        return cls.NO_POSITION


def add_to_context_menu(
    app_name: str,
    path: str | os.PathLike[str],
    icon_path: str | os.PathLike[str] | None = None,
):
    import winreg

    key = winreg.HKEY_CURRENT_USER
    handle = winreg.CreateKeyEx(
        key,
        f"Software\\Classes\\*\\shell\\Open with {app_name}\\command",
        0,
        winreg.KEY_SET_VALUE,
    )
    if getattr(sys, "frozen", False):
        command_value = f"{path} %1"
    else:
        command_value = rf'{sys.executable} -m {path} "%V"'
    winreg.SetValueEx(handle, "", 0, winreg.REG_SZ, command_value)
    handle = winreg.CreateKeyEx(
        key, f"Software\\Classes\\*\\shell\\Open with {app_name}", 0, winreg.KEY_SET_VALUE
    )
    if icon_path:
        winreg.SetValueEx(handle, "icon", 0, winreg.REG_SZ, os.fspath(icon_path))


def remove_from_context_menu(self, app_name: str):
    import winreg

    key = winreg.HKEY_CURRENT_USER
    winreg.DeleteKey(key, f"Software\\Classes\\*\\shell\\Open with {app_name}\\command")
    winreg.DeleteKey(key, f"Software\\Classes\\*\\shell\\Open with {app_name}")



from __future__ import annotations

from prettyqt import core, widgets


class OptionalWidget(widgets.GroupBox):
    """Wraps another widget in a GroupBox with and CheckBock and makes it optional."""

    value_changed = core.Signal(object)

    def __init__(
        self,
        widget: widgets.QWidget,
        title: str = "",
        parent: widgets.QWidget | None = None,
    ):
        super().__init__(checkable=True, title=title)
        self.set_layout("vertical")
        self.box.add(widget)
        self.widget = widget
        self.toggled.connect(self.widget.setEnabled)
        self.widget.value_changed.connect(self._on_value_change)
        self.toggled.connect(self._on_enable_change)

    def _on_value_change(self, val):
        self.value_changed.emit(val)

    def _on_enable_change(self, val):
        self.widget.setEnabled(val)
        if val:
            self.value_changed.emit(self.widget.get_value())
        else:
            self.value_changed.emit(None)

    def __getattr__(self, value: str):
        return getattr(self.widget, value)

    @classmethod
    def setup_example(cls):
        w = widgets.CheckBox("Example")
        return cls(widget=w)

    @property
    def enabled(self) -> bool:
        return self.isChecked()

    @enabled.setter
    def enabled(self, state: bool):
        self.setChecked(state)

    def get_value(self):
        return self.widget.get_value() if self.isChecked() else None


if __name__ == "__main__":
    app = widgets.app()
    img = widgets.RadioButton("test")
    widget = OptionalWidget(img, "Test")
    widget.value_changed.connect(print)
    widget.show()
    app.exec()



from __future__ import annotations

from dataclasses import dataclass
import enum
from typing import Literal

from prettyqt import constants, core, widgets


@dataclass
class ItemWrapper:
    item: widgets.QWidgetItem
    position: BorderLayout.Position


class BorderLayout(widgets.Layout):
    """Layout based on 5 Regions (north east, south, west and center)."""

    ID = "border"

    class Position(enum.IntEnum):
        """Item position."""

        West = 0
        North = 1
        South = 2
        East = 3
        Center = 4

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items: list[ItemWrapper] = []

    def addItem(self, item: widgets.QWidgetItem):
        self.add_widgetitem(item, BorderLayout.Position.West)

    def addWidget(
        self,
        widget: widgets.QWidget,
        position: Position | None = None,
    ):
        position = BorderLayout.Position.West if position is None else position
        self.add_widgetitem(widgets.WidgetItem(widget), position)

    def expandingDirections(self) -> constants.Orientation:
        return constants.HORIZONTAL | constants.VERTICAL

    def hasHeightForWidth(self) -> bool:
        return False

    def count(self) -> int:
        return len(self.items)

    def itemAt(self, index: int) -> widgets.QWidgetItem | None:
        return self.items[index].item if index < len(self.items) else None

    def minimumSize(self):
        return self.calculate_size("minimum")

    def setGeometry(self, rect: core.QRect):
        center = None
        east_width = 0
        west_width = 0
        north_height = 0
        south_height = 0
        super().setGeometry(rect)
        for wrapper in self.items:
            item = wrapper.item
            match wrapper.position:
                case BorderLayout.Position.North:
                    h = item.sizeHint().height()
                    geom = core.Rect(rect.x(), north_height, rect.width(), h)
                    item.setGeometry(geom)
                    north_height += item.geometry().height() + self.spacing()
                case BorderLayout.Position.South:
                    geo = item.geometry()
                    h = item.sizeHint().height()
                    geom = core.Rect(geo.x(), geo.y(), rect.width(), h)
                    item.setGeometry(geom)
                    south_height += item.geometry().height() + self.spacing()
                    y = rect.y() + rect.height() - south_height + self.spacing()
                    geo = item.geometry()
                    geom = core.Rect(rect.x(), y, geo.width(), geo.height())
                    item.setGeometry(geom)
                case BorderLayout.Position.Center:
                    center = wrapper

        center_height = rect.height() - north_height - south_height

        for wrapper in self.items:
            item = wrapper.item
            match wrapper.position:
                case BorderLayout.Position.West:
                    x = rect.x() + west_width
                    w = item.sizeHint().width()
                    geom = core.Rect(x, north_height, w, center_height)
                    item.setGeometry(geom)
                    west_width += item.geometry().width() + self.spacing()
                case BorderLayout.Position.East:
                    geo = item.geometry()
                    w = item.sizeHint().width()
                    geom = core.Rect(geo.x(), geo.y(), w, center_height)
                    item.setGeometry(geom)
                    east_width += item.geometry().width() + self.spacing()
                    x = rect.x() + rect.width() - east_width + self.spacing()
                    geom = core.Rect(x, north_height, geo.width(), geo.height())
                    item.setGeometry(geom)

        if center:
            w = rect.width() - east_width - west_width
            rect = core.Rect(west_width, north_height, w, center_height)
            center.item.setGeometry(rect)

    def sizeHint(self) -> core.Size:
        return self.calculate_size("size_hint")

    def takeAt(self, index: int) -> widgets.QWidgetItem | None:
        if 0 <= index < len(self.items):
            layout_struct = self.items.pop(index)
            return layout_struct.item

        return None

    def add_widgetitem(self, item: widgets.QWidgetItem, position: Position):
        self.items.append(ItemWrapper(item, position))

    def calculate_size(self, size_type: Literal["minimum", "size_hint"]) -> core.Size:
        total_size = core.Size()
        pos = BorderLayout.Position
        for wrapper in self.items:
            item_size = (
                wrapper.item.minimumSize()
                if size_type == "minimum"
                else wrapper.item.sizeHint()
            )
            if wrapper.position in (pos.North, pos.South, pos.Center):
                total_size.setHeight(total_size.height() + item_size.height())
            if wrapper.position in (pos.West, pos.East, pos.Center):
                total_size.setWidth(total_size.width() + item_size.width())
        return total_size


if __name__ == "__main__":

    class Window(widgets.Widget):
        def __init__(self):
            super().__init__()

            central_widget = widgets.TextBrowser()
            central_widget.setPlainText("Central widget")

            layout = BorderLayout()
            layout.addWidget(central_widget, BorderLayout.Position.Center)

            # Because BorderLayout doesn't call its super-class addWidget() it
            # doesn't take ownership of the widgets until setLayout() is called.
            # Therefore we keep a local reference to each label to prevent it being
            # garbage collected too soon.
            label_n = self.create_label("North")
            layout.addWidget(label_n, BorderLayout.Position.North)
            label_w = self.create_label("West")
            layout.addWidget(label_w, BorderLayout.Position.West)
            label_e1 = self.create_label("East 1")
            layout.addWidget(label_e1, BorderLayout.Position.East)
            label_e2 = self.create_label("East 2")
            layout.addWidget(label_e2, BorderLayout.Position.East)
            label_s = self.create_label("South")
            layout.addWidget(label_s, BorderLayout.Position.South)
            self.setLayout(layout)
            self.setWindowTitle("Border Layout")

        @staticmethod
        def create_label(text: str):
            label = widgets.Label(text)
            label.setFrameStyle(widgets.Frame.Shape.Box | widgets.Frame.Shadow.Raised)
            return label

    app = widgets.app()
    window = Window()
    window.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import location, positioning


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class GeoRouteMixin:
    def __setitem__(self, index: str, val: datatypes.Variant):
        attrs = self.extendedAttributes()
        attrs[index] = val
        self.setExtendedAttributes(attrs)

    def __getitem__(self, index: str) -> datatypes.Variant:
        attr = self.extendedAttributes()
        if index not in attr:
            msg = f"Key {index!r} does not exist."
            raise KeyError(msg)
        return attr[index]

    def __abs__(self) -> float:
        return self.distance()

    def get_bounds(self) -> positioning.GeoRectangle:
        return positioning.GeoRectangle(self.bounds())

    def get_first_route_segment(self) -> location.GeoRouteSegment:
        return location.GeoRouteSegment(self.firstRouteSegment())

    def get_path(self) -> list[positioning.GeoCoordinate]:
        return [positioning.GeoCoordinate(i) for i in self.path()]


class GeoRoute(GeoRouteMixin, location.QGeoRoute):
    """Represents a route between two points."""



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtStateMachine


class AbstractStateMixin(core.ObjectMixin):
    pass


class AbstractState(AbstractStateMixin, QtStateMachine.QAbstractState):
    pass


if __name__ == "__main__":
    state = AbstractState()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import constants, widgets
from prettyqt.utils import helpers, listdelegators


if TYPE_CHECKING:
    from collections.abc import Iterator


class GridLayout(widgets.LayoutMixin, widgets.QGridLayout):
    """Lays out widgets in a grid."""

    ID = "grid"

    def __getitem__(
        self, index: tuple[int | slice, int | slice] | int | str
    ) -> (
        widgets.QWidget
        | widgets.QLayout
        | listdelegators.ListDelegator[widgets.QWidget | widgets.QLayout]
        | None
    ):
        rowcount = self.rowCount()
        colcount = self.columnCount()
        match index:
            case int() as row, int() as col:
                if row < 0:
                    row = rowcount + row
                if col < 0:
                    col = colcount + col
                if row >= rowcount or row < 0 or col >= colcount or col < 0:
                    raise IndexError(index)
                item = self.itemAtPosition(row, col)
                return i if (i := item.widget()) is not None else item.layout()
            case (row, col):
                items = [
                    item
                    for i, j in helpers.iter_positions(row, col, rowcount, colcount)
                    if (item := self.itemAtPosition(i, j)) is not None
                ]
                items = [
                    w if (w := i.widget()) is not None else i.layout() for i in items
                ]
                return listdelegators.ListDelegator(list(set(items)))
            case int() as row:
                if row < 0:
                    row = rowcount + row
                if row >= rowcount or row < 0:
                    raise IndexError(index)
                item = self.itemAt(row)
                return i if (i := item.widget()) is not None else item.layout()
            case slice() as rowslice:
                count = rowcount if rowslice.stop is None else rowslice.stop
                items = [self.itemAt(i) for i in range(count)[rowslice]]
                items = [
                    w if (w := i.widget()) is not None else i.layout() for i in items
                ]
                return listdelegators.ListDelegator(list(set(items)))
            case str():
                return self.find_child(widgets.QWidget | widgets.QLayout, index)
            case _:
                raise TypeError(index)

    def __setitem__(
        self,
        idx: tuple[int | slice, int | slice],
        value: widgets.QWidget | widgets.QLayout | widgets.QLayoutItem,
    ):
        row, col = idx
        rowspan = row.stop - row.start + 1 if isinstance(row, slice) else 1
        colspan = col.stop - col.start + 1 if isinstance(col, slice) else 1
        rowstart = row.start if isinstance(row, slice) else row
        colstart = col.start if isinstance(col, slice) else col
        self.add(value, rowstart, colstart, rowspan, colspan)

    # def __setstate__(self, state):
    #     for item, pos in zip(state["widgets"], state["positions"]):
    #         x, y, w, h = pos
    #         self[x : x + w - 1, y : y + h - 1] = item

    # def __reduce__(self):
    #     return type(self), (), self.__getstate__()

    def __iter__(self) -> Iterator[widgets.QWidget | widgets.QLayout]:
        return iter(
            item for i in range(self.count()) if (item := self.itemAt(i)) is not None
        )

    def __add__(
        self,
        other: (tuple | list | widgets.QWidget | widgets.QLayout | widgets.QLayoutItem),
    ):
        if isinstance(other, tuple | list):
            for i, _control in enumerate(other):
                self[self.rowCount(), i] = other  # type: ignore
        else:
            self[self.rowCount(), 0 : self.columnCount() - 1] = other
        return self

    def __iadd__(self, item, *args, **kwargs):
        self.__add__(item, *args, **kwargs)
        return self

    def add(
        self,
        item: widgets.QWidget | widgets.QLayout | widgets.QLayoutItem,
        rowstart: int,
        colstart: int,
        rowspan: int = 1,
        colspan: int = 1,
        alignment: constants.AlignmentStr | constants.AlignmentFlag | None = None,
    ):
        if alignment is None:
            alignment = "none"
        flag = constants.ALIGNMENTS.get_enum_value(alignment)
        match item:
            case widgets.QWidget():
                self.addWidget(item, rowstart, colstart, rowspan, colspan, flag)
            case widgets.QLayout():
                self.addLayout(item, rowstart, colstart, rowspan, colspan, flag)
            case widgets.QLayoutItem():
                self.addItem(item, rowstart, colstart, rowspan, colspan, flag)

    def append(self, item: widgets.QWidget | widgets.QLayout | widgets.QLayoutItem):
        self[self.rowCount(), 0 : self.columnCount() - 1] = item

    def set_origin_corner(self, corner: constants.CornerStr | constants.Corner):
        """Set the origin corner.

        Args:
            corner: origin corner
        """
        self.setOriginCorner(constants.CORNER.get_enum_value(corner))

    def get_origin_corner(self) -> constants.CornerStr:
        """Return current origin corner.

        Returns:
            origin corner
        """
        return constants.CORNER.inverse[self.originCorner()]


if __name__ == "__main__":
    app = widgets.app()
    layout = GridLayout()
    layout[1, 5:6] = widgets.RadioButton("1 2 3 jk jkjl j kföldsjfköj")
    layout[3:5, 7:8] = widgets.RadioButton("2")
    layout[3:5, 1:4] = widgets.RadioButton("3")
    layout += widgets.RadioButton("3")
    layout += widgets.RadioButton("4")
    widget = widgets.Widget()
    widget.set_layout(layout)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import gui
from prettyqt.utils import bidict


PositionStr = Literal["in_flow", "flow_right", "flow_left"]

POSITIONS: bidict[PositionStr, gui.QTextFrameFormat.Position] = bidict(
    in_flow=gui.QTextFrameFormat.Position.InFlow,
    flow_left=gui.QTextFrameFormat.Position.FloatLeft,
    flow_right=gui.QTextFrameFormat.Position.FloatRight,
)

BorderStyleStr = Literal[
    "none",
    "dotted",
    "dashed",
    "solid",
    "double",
    "dot_dash",
    "dot_dot_dash",
    "groove",
    "ridge",
    "inset",
    "outset",
]

BORDER_STYLES: bidict[BorderStyleStr, gui.QTextFrameFormat.BorderStyle] = bidict(
    none=gui.QTextFrameFormat.BorderStyle.BorderStyle_None,
    dotted=gui.QTextFrameFormat.BorderStyle.BorderStyle_Dotted,
    dashed=gui.QTextFrameFormat.BorderStyle.BorderStyle_Dashed,
    solid=gui.QTextFrameFormat.BorderStyle.BorderStyle_Solid,
    double=gui.QTextFrameFormat.BorderStyle.BorderStyle_Double,
    dot_dash=gui.QTextFrameFormat.BorderStyle.BorderStyle_DotDash,
    dot_dot_dash=gui.QTextFrameFormat.BorderStyle.BorderStyle_DotDotDash,
    groove=gui.QTextFrameFormat.BorderStyle.BorderStyle_Groove,
    ridge=gui.QTextFrameFormat.BorderStyle.BorderStyle_Ridge,
    inset=gui.QTextFrameFormat.BorderStyle.BorderStyle_Inset,
    outset=gui.QTextFrameFormat.BorderStyle.BorderStyle_Outset,
)


class TextFrameFormatMixin(gui.TextFormatMixin):
    def get_height(self) -> gui.TextLength:
        length = self.height()
        return gui.TextLength(length.type(), length.rawValue())

    def get_width(self) -> gui.TextLength:
        length = self.width()
        return gui.TextLength(length.type(), length.rawValue())

    def get_border_brush(self) -> gui.Brush:
        return gui.Brush(self.borderBrush())

    def set_border_style(self, style: BorderStyleStr | gui.QTextFrameFormat.BorderStyle):
        """Set border style.

        Args:
            style: border style
        """
        self.setBorderStyle(BORDER_STYLES.get_enum_value(style))

    def get_border_style(self) -> BorderStyleStr:
        """Get the current border style.

        Returns:
            border style
        """
        return BORDER_STYLES.inverse[self.borderStyle()]

    def set_page_break_policy(
        self, policy: gui.textformat.PageBreakFlagStr | gui.QTextFormat.PageBreakFlag
    ):
        """Set page break policy.

        Args:
            policy: page break policy
        """
        self.setPageBreakPolicy(gui.textformat.PAGE_BREAK_FLAG.get_enum_value(policy))

    def get_page_break_policy(self) -> gui.textformat.PageBreakFlagStr:
        """Get the current page break policy.

        Returns:
            page break policy
        """
        return gui.textformat.PAGE_BREAK_FLAG.inverse[self.pageBreakPolicy()]

    def set_position(self, position: PositionStr | gui.QTextFrameFormat.Position):
        """Set position.

        Args:
            position: position
        """
        self.setPosition(POSITIONS.get_enum_value(position))

    def get_position(self) -> PositionStr:
        """Get the current position.

        Returns:
            position
        """
        return POSITIONS.inverse[self.position()]


class TextFrameFormat(TextFrameFormatMixin, gui.QTextFrameFormat):
    pass



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict


FlagStr = Literal["hidden_from_help", "short_option_style"]

FLAGS: bidict[FlagStr, core.QCommandLineOption.Flag] = bidict(
    hidden_from_help=core.QCommandLineOption.Flag.HiddenFromHelp,
    short_option_style=core.QCommandLineOption.Flag.ShortOptionStyle,
)


class CommandLineOption(core.QCommandLineOption):
    """Defines a possible command-line option."""



from __future__ import annotations

import logging

from prettyqt import constants, core, eventfilters, widgets


logger = logging.getLogger(__name__)


class DebugMode(eventfilters.BaseEventFilter):
    ID = "debug_mode"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.menu = widgets.Menu(
            hovered=self._on_entered,
            triggered=self._on_clicked,
            about_to_hide=self._on_hide,
        )
        self.frame = widgets.RubberBand("rectangle")
        self.frame.setObjectName("testframe")

    def eventFilter(self, source: core.QObject, event: core.QEvent) -> bool:
        border_keys = [constants.Key.Key_Control, constants.Key.Key_Alt]
        match event.type():
            case core.Event.Type.KeyPress if event.key() in border_keys:
                with widgets.app().edit_stylesheet() as ss:
                    ss.QWidget.border.setValue("1px solid red")
            case core.Event.Type.KeyRelease if event.key() in border_keys:
                widgets.app().setStyleSheet("")
            case core.Event.Type.MouseButtonPress:
                mods = widgets.Application.query_keyboard_modifiers()
                if "ctrl" in mods:
                    pos = source.mapToGlobal(event.pos())
                    candidates = widgets.Application.widgets_at(pos)
                    self.menu.clear()
                    for candidate in candidates:
                        self.menu.add_action(text=repr(candidate), data=candidate)
                    self.menu.exec(pos)
                    return True
                if "alt" in mods:
                    from prettyqt import ipython

                    self.frame.hide()
                    console = ipython.InProcessIPythonWidget(self)
                    console.show()
                    pos = source.mapToGlobal(event.pos())
                    widgets.Application.sleep(1)
                    variables = {}
                    i = 0
                    for w in widgets.Application.widgets_at(pos):
                        if (name := w.objectName()) and name not in variables:
                            variables[name] = w
                        else:
                            name = f"{type(w).__name__}_{i}".lower()
                            variables[name] = w
                            i += 1
                    console.push_vars(dict(app=widgets.app(), **variables))
            # case core.QEvent.Type.ToolTip if source.isWidgetType():
            #     metaobj = core.MetaObject(source.metaObject())
            #     lines = [
            #         f"{k}: {v!r}"
            #         for k, v in metaobj.get_property_values(source).items()
            #         if not k == "toolTip"
            #     ]
            #     source.setToolTip("<br>".join(lines))

        return False

    def _on_hide(self):
        self.frame.hide()

    def _on_clicked(self, item):
        from prettyqt import debugging

        logger.debug("clicked on %s", item)
        widget = item.data()
        self.editor = debugging.QObjectDetailsDialog(widget)
        self.menu.close()
        # self.frame.hide()
        self.editor.show()

    def _on_entered(self, index):
        widget = index.data()
        # logger.info(widget)
        for item in self.menu:
            if item.data() == widget:
                self.frame.resize(widget.size())
                # self.frame.setParent(widget)
                pos = widget.mapToGlobal(core.Point(0, 0))
                self.frame.move(pos)
                self.menu.raise_()
                self.frame.show()


if __name__ == "__main__":
    app = widgets.app()
    container = widgets.Widget()
    container.set_layout("horizontal")
    w = widgets.PlainTextEdit(parent=container)
    w2 = widgets.RadioButton(parent=container)
    container.box.add(w)
    container.box.add(w2)
    container.show()
    with app.debug_mode():
        app.sleep(1)
        app.exec()



from __future__ import annotations

from prettyqt import constants, core, gui


class Screen:
    """Used to query screen properties."""

    def __init__(self, item: gui.QScreen):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_geometry(self) -> core.Rect:
        return core.Rect(self.geometry())

    def get_size(self) -> core.Size:
        return core.Size(self.size())

    def get_available_geometry(self) -> core.Rect:
        return core.Rect(self.availableGeometry())

    def get_available_size(self) -> core.Size:
        return core.Size(self.availableSize())

    def get_available_virtual_geometry(self) -> core.Rect:
        return core.Rect(self.availableVirtualGeometry())

    def get_available_virtual_size(self) -> core.Size:
        return core.Size(self.availableVirtualSize())

    def get_virtual_geometry(self) -> core.Rect:
        return core.Rect(self.virtualGeometry())

    def get_virtual_size(self) -> core.Size:
        return core.Size(self.virtualSize())

    def get_native_orientation(self) -> constants.ScreenOrientationStr:
        return constants.SCREEN_ORIENTATION.inverse[self.nativeOrientation()]

    def get_orientation(self) -> constants.ScreenOrientationStr:
        return constants.SCREEN_ORIENTATION.inverse[self.orientation()]

    def get_primary_orientation(self) -> constants.ScreenOrientationStr:
        return constants.SCREEN_ORIENTATION.inverse[self.primaryOrientation()]

    def get_physical_size(self) -> core.SizeF:
        return core.SizeF(self.physicalSize())

    def get_angle_between(
        self,
        orientation_1: constants.ScreenOrientationStr | constants.ScreenOrientation,
        orientation_2: constants.ScreenOrientationStr | constants.ScreenOrientation,
    ):
        self.angleBetween(
            constants.SCREEN_ORIENTATION.get_enum_value(orientation_1),
            constants.SCREEN_ORIENTATION.get_enum_value(orientation_2),
        )

    def grab_window(
        self,
        window: int = 0,
        x: int = 0,
        y: int = 0,
        width: int | None = None,
        height: int | None = None,
    ) -> gui.Pixmap:
        if width is None:
            width = -1
        if height is None:
            height = -1
        px = self.grabWindow(window, x, y, width, height)
        return gui.Pixmap(px)

    def get_virtual_siblings(self) -> list[Screen]:
        return [Screen(i) for i in self.virtualSiblings()]



from __future__ import annotations

import enum
import logging
from typing import Literal

from prettyqt import constants, core, gui, itemmodels, widgets
from prettyqt.utils import bidict


logger = logging.getLogger(__name__)


class SectionWidget(widgets.Widget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_layout("vertical", margin=0, spacing=0)
        self.set_margin(0)
        header = widgets.Label(self.windowTitle()).set_bold()
        header.set_point_size(header.font().pointSize() * 2)
        self.box.add(header)
        for i in range(5):
            subitem = widgets.CheckBox(f"sub item {i}")
            self.box.add(subitem)


class ScrollAreaTocModel(itemmodels.TreeModel):
    def __init__(
        self, *args, widget_class: type[widgets.QWidget] = widgets.QWidget, **kwargs
    ):
        self._highlight_font = gui.QFont()
        self._highlight_font.setBold(True)
        self._Class = widget_class
        self._current_widgets: list[widgets.QWidget] = []
        self._header_property = "windowTitle"
        super().__init__(*args, **kwargs)

    def set_highlighted_widgets(self, widgets: list[widgets.QWidget]):
        self._current_widgets = widgets
        self.update_all()

    def columnCount(self, parent=None):
        return 1

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        widget = self.data_by_index(index).obj
        match role:
            case constants.DISPLAY_ROLE if widget in self._current_widgets:
                return f"> {widget.property(self.header_property)}"
            case constants.DISPLAY_ROLE:
                return widget.property(self.header_property)
            case constants.DECORATION_ROLE:
                return widget.windowIcon()
            case constants.USER_ROLE:
                return widget
            case constants.FONT_ROLE if widget in self._current_widgets:
                # if not index.parent().isValid():
                return self._highlight_font

    def _fetch_object_children(
        self, item: ScrollAreaTocModel.TreeItem
    ) -> list[ScrollAreaTocModel.TreeItem]:
        flag = constants.FindChildOption.FindDirectChildrenOnly
        children = item.obj.findChildren(self._Class, None, flag)
        children = [i for i in children if i.property(self.header_property)]
        return [self.TreeItem(obj=i) for i in children]

    def _has_children(self, item: ScrollAreaTocModel.TreeItem) -> bool:
        flag = constants.FindChildOption.FindDirectChildrenOnly
        children = item.obj.findChildren(self._Class, None, flag)
        children = [i for i in children if i.property(self.header_property)]
        return bool(children)

    def set_highlight_font(self, font: gui.QFont):
        self._highlight_font = font

    def get_highlight_font(self) -> gui.QFont:
        return self._highlight_font

    def set_header_property(self, prop: str):
        self._header_property = prop

    def get_header_property(self) -> str:
        return self._header_property

    highlight_font = core.Property(
        gui.QFont,
        get_highlight_font,
        set_highlight_font,
        doc="Font to use for highlighting",
    )
    header_property = core.Property(
        str,
        get_header_property,
        set_header_property,
        doc="Property to use for the list labels",
    )


class ScrollMode(enum.Enum):
    Single = 1
    Multi = 2
    HeadersOnly = 4


ScrollModeStr = Literal["single", "multi", "headers_only"]

SCROLL_MODE: bidict[ScrollModeStr, ScrollMode] = bidict(
    single=ScrollMode.Single, multi=ScrollMode.Multi, headers_only=ScrollMode.HeadersOnly
)


class ScrollAreaTocWidget(widgets.TreeView):
    section_changed = core.Signal()

    ScrollMode = ScrollMode
    core.Enum(ScrollMode)

    def __init__(
        self,
        scrollarea: widgets.QScrollArea,
        orientation: (
            constants.Orientation | constants.OrientationStr
        ) = constants.VERTICAL,
        widget_class: type = widgets.QWidget,
        **kwargs,
    ) -> None:
        """A TreeView which can show a Table-of-contents list based on a given ScrollArea.

        With default settings, it will scan the widgets contained in the scrollArea for
         a windowTitle.
        These widgets will be shown in the TreeView, and the TreeView selection will be
        synced to what is currently visible in the ScrollArea.
        This basically emulates the behaviour of many websites
        (like the [Qt Website][https://doc.qt.io/qt-6/supported-platforms.html])
        or from VS code settings.

        You can set up the Toc Tree by passing a ScrollArea instance:

        ```py
        widget = widgets.Widget()
        layout = widget.set_layout("horizontal")
        scrollarea = widgets.ScrollArea()
        # add some widgets to the ScrollArea here
        # ...
        toc = ScrollAreaTocWidget(scrollarea)
        layout.add(toc)
        layout.add(scrollarea)
        ```

        The Toc tree can be configured to use other properties for populating
        than windowTitle.
        You can also set a widget class filter if you only want a specific widget
        class to be shown.

        The widget supports 3 different scroll modes:

        1) Single: Only one row in the Tree will be highlighted. (the one which equals
        the topmost one from the scrollArea)
        2) Multi: All rows which equal to visible widgets in the ScrollArea will be shown.
        3) HeadersOnly: like Single, but only highlights top-level widgets from the
                        ScrollArea.

        There are also two different expand modes to choose from:

        1) ExpandAll: All subsections are always expanded.
        2) on_focus: Only the section containing the focused item is expanded.

        This widget is used by [ConfigWidget](ConfigWidget.md).
        """
        # TODO: not sure if parent should always equal scrollarea..."""
        self._WidgetClass = widget_class
        self._scroll_mode = ScrollMode.Single
        self._always_expanded = False
        self._last_visible = None
        self.scrollarea = scrollarea
        super().__init__(scrollarea, **kwargs)
        self._orientation = constants.ORIENTATION.get_enum_value(orientation)
        self.setFixedWidth(200)
        self.h_header.hide()
        self.h_header.setStretchLastSection(True)
        self.setAlternatingRowColors(False)
        self.setRootIsDecorated(False)
        # self.setStyleSheet(
        #     """::item:hover {background: transparent; border-color:transparent}
        #     ::item:selected { border-color:transparent;
        #     border-style:outset; border-width:2px; color:black; }"""
        # )
        if self._orientation == constants.VERTICAL:
            scrollarea.v_scrollbar.valueChanged.connect(self._on_scroll)
        else:
            scrollarea.h_scrollbar.valueChanged.connect(self._on_scroll)
        self.set_widget(scrollarea)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"scrollMode": SCROLL_MODE}
        return maps

    def showEvent(self, event):
        super().showEvent(event)
        self._on_scroll()

    # @classmethod
    # def setup_example(cls):
    #     scrollarea = widgets.ScrollArea()
    #     w = widgets.Widget()
    #     scrollarea.set_widget(w)
    #     return cls(scrollarea)

    def set_widget(self, widget: widgets.QScrollArea):
        """Set the ScrollArea widget to follow."""
        if widget.widget() is None:
            msg = "No widget set on ScrollArea."
            raise RuntimeError(msg)
        self.scrollarea = widget
        model = ScrollAreaTocModel(
            widget.widget(),
            show_root=True,
            parent=self.scrollarea,
            widget_class=self._WidgetClass,
        )
        self.set_model(model)
        self.proxy = self.proxifier.set_sort_filter_proxy(
            recursive_filtering_enabled=True
        )
        self.proxy.set_filter_case_sensitive(False)
        self.show_root(False)
        widget.widget().installEventFilter(self)
        self.selectionModel().currentRowChanged.connect(self._on_current_change)
        self.selectionModel().selectionChanged.connect(self._on_selection_change)
        # if self._always_expanded:
        self.expandAll()

    def _on_current_change(self, new, old):
        if self.model() is None:
            return
        is_vertical = self._orientation == constants.VERTICAL
        area = self.scrollarea
        scrollbar = area.v_scrollbar if is_vertical else area.h_scrollbar
        with self.signal_blocked(scrollbar.valueChanged, self._on_scroll):
            widget = self.model().data(new, role=constants.USER_ROLE)
            area.scroll_to_bottom()
            area.ensureWidgetVisible(widget, 10, 10)

    def _on_selection_change(self, new, old):
        if self.model() is None:
            return
        widgets = [i.data(constants.USER_ROLE) for i in self.selected_indexes()]
        self.get_model(skip_proxies=True).set_highlighted_widgets(widgets)

    def _on_scroll(self):
        model: ScrollAreaTocModel | None = self.model()
        if model is None:
            return
        visible_widgets = self.scrollarea.get_visible_widgets(typ=self._WidgetClass)
        if not visible_widgets or visible_widgets == self._last_visible:
            return
        self._last_visible = visible_widgets
        sig = self.selectionModel().currentRowChanged
        with self.signal_blocked(sig, self._on_current_change):
            self.select_index(None)
            if not self._always_expanded:
                self.collapseAll()
            match self.get_scroll_mode():
                case "multi":
                    indexes = model.search_tree(visible_widgets, constants.USER_ROLE)
                    for index in indexes:
                        children = model.get_child_indexes(index)
                        # only select if all children selected.
                        # if all(c in indexes for c in children):

                        # highlight when no children or when first child is visible.
                        if not children or children[0] in indexes:
                            self.select_index(index, clear=False)
                        self.set_expanded(indexes)
                        self.scroll_to(indexes[0])
                        self.scroll_to(indexes[-1])
                case "headers_only":
                    if indexes := model.search_tree(
                        visible_widgets,
                        role=constants.USER_ROLE,
                        max_results=1,
                    ):
                        self.set_current_index(indexes[0])
                        self.scroll_to(indexes[0])
                case "single":
                    if indexes := model.search_tree(visible_widgets, constants.USER_ROLE):
                        viewport = self.scrollarea.viewport()
                        # sort indexes by closest distance to top
                        indexes.sort(
                            key=lambda x: abs(
                                x.data(constants.USER_ROLE)
                                .map_to(viewport, x.data(constants.USER_ROLE).rect())
                                .top()
                            ),
                        )
                        self.collapseAll()
                        self.model().fetchMore(indexes[0])
                        self.set_current_index(indexes[0])
                        self.scroll_to(indexes[0])
                case _:
                    raise ValueError(self._scroll_mode)
        # model.set_highlighted_indexes(indexes)

    def wheelEvent(self, e):
        self.scrollarea.wheelEvent(e)

    def eventFilter(self, source: core.QObject, event: core.QEvent) -> bool:
        match event.type():
            case core.Event.Type.ChildAdded:
                self._on_scroll()
        return False

    def _scrollMode(self) -> ScrollMode:
        return self._scroll_mode

    def get_scroll_mode(self) -> ScrollModeStr:
        return SCROLL_MODE.inverse[self._scrollMode()]

    def set_scroll_mode(self, mode: ScrollMode | ScrollModeStr):
        self._scroll_mode = SCROLL_MODE.get_enum_value(mode)

    def is_always_expanded(self) -> bool:
        return self._always_expanded

    def set_always_expanded(self, always_expanded: bool):
        self._always_expanded = always_expanded

    scrollMode = core.Property(  # noqa: N815
        enum.Enum,
        _scrollMode,
        set_scroll_mode,
        doc="Scrolling mode",
    )
    always_expanded = core.Property(
        bool,
        is_always_expanded,
        set_always_expanded,
        doc="Whether the tree is always expanded.",
    )


if __name__ == "__main__":
    app = widgets.app()
    with app.debug_mode():
        window = widgets.Widget(object_name="window")
        layout = window.set_layout("horizontal")
        scrollarea = widgets.ScrollArea(
            object_name="scroll", parent=window, widget_resizable=True
        )
        widget = widgets.Widget(object_name="scrollarea_widget")
        scrollarea.setWidget(widget)
        toc_widget = ScrollAreaTocWidget(
            scrollarea,
            scroll_mode="headers_only",
            widget_class=SectionWidget,
        )
        layout.add(toc_widget)
        scroll_layout = widget.set_layout("vertical")
        layout.add(scrollarea)
        for i in range(10):
            section = SectionWidget(window_title=f"test{i}")
            section.box.add(SectionWidget(window_title=f"{i}nested"))
            section.box.add(SectionWidget(window_title=f"{i * 10}nested"))
            scroll_layout.add(section)
        window.show()
        app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core, gui


if TYPE_CHECKING:
    from prettyqt.qt import QtQuick


class QuickItemGrabResult(core.ObjectMixin):
    """Contains the result from QQuickItem::grabToImage()."""

    def __init__(self, item: QtQuick.QQuickItemGrabResult):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_image(self) -> gui.Image:
        return gui.Image(self.image())

    def get_url(self) -> core.Url:
        return core.Url(self.url())


# if __name__ == "__main__":
#     item = QuickItemGrabResult()



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING, Any, Literal, Self

from prettyqt.qt import QtCore
from prettyqt.utils import bidict, get_repr, serializemixin


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


ComponentFormattingStr = Literal[
    "pretty_decoded",
    "encode_spaces",
    "encode_unicode",
    "encode_delimiters",
    "encode_reserved",
    "decode_reserved",
    "fully_encoded",
    "fully_decoded",
]

COMPONENT_FORMATTING_OPTIONS: bidict[
    ComponentFormattingStr, QtCore.QUrl.ComponentFormattingOption
] = bidict(
    pretty_decoded=QtCore.QUrl.ComponentFormattingOption.PrettyDecoded,
    encode_spaces=QtCore.QUrl.ComponentFormattingOption.EncodeSpaces,
    encode_unicode=QtCore.QUrl.ComponentFormattingOption.EncodeUnicode,
    encode_delimiters=QtCore.QUrl.ComponentFormattingOption.EncodeDelimiters,
    encode_reserved=QtCore.QUrl.ComponentFormattingOption.EncodeReserved,
    decode_reserved=QtCore.QUrl.ComponentFormattingOption.DecodeReserved,
    fully_encoded=QtCore.QUrl.ComponentFormattingOption.FullyEncoded,
    fully_decoded=QtCore.QUrl.ComponentFormattingOption.FullyDecoded,
)

ParsingModeStr = Literal["tolerant", "strict", "decoded"]

PARSING_MODES: bidict[FormattingOptionStr, QtCore.QUrl.ParsingMode] = bidict(
    tolerant=QtCore.QUrl.ParsingMode.TolerantMode,
    strict=QtCore.QUrl.ParsingMode.StrictMode,
    decoded=QtCore.QUrl.ParsingMode.DecodedMode,
)

FormattingOptionStr = Literal[
    "none",
    "remove_scheme",
    "remove_password",
    "remove_user_info",
    "remove_port",
    "remove_authority",
    "remove_path",
    "remove_query",
    "remove_fragment",
    "remove_filename",
    "prefer_local_file",
    "strip_trailing_slash",
    "normalize_path_segments",
]

FORMATTING_OPTIONS: bidict[FormattingOptionStr, QtCore.QUrl.UrlFormattingOption] = bidict(
    none=0,  # QtCore.QUrl.UrlFormattingOption.None
    remove_scheme=QtCore.QUrl.UrlFormattingOption.RemoveScheme,
    remove_password=QtCore.QUrl.UrlFormattingOption.RemovePassword,
    remove_user_info=QtCore.QUrl.UrlFormattingOption.RemoveUserInfo,
    remove_port=QtCore.QUrl.UrlFormattingOption.RemovePort,
    remove_authority=QtCore.QUrl.UrlFormattingOption.RemoveAuthority,
    remove_path=QtCore.QUrl.UrlFormattingOption.RemovePath,
    remove_query=QtCore.QUrl.UrlFormattingOption.RemoveQuery,
    remove_fragment=QtCore.QUrl.UrlFormattingOption.RemoveFragment,
    remove_filename=QtCore.QUrl.UrlFormattingOption.RemoveFilename,
    prefer_local_file=QtCore.QUrl.UrlFormattingOption.PreferLocalFile,
    strip_trailing_slash=QtCore.QUrl.UrlFormattingOption.StripTrailingSlash,
    normalize_path_segments=QtCore.QUrl.UrlFormattingOption.NormalizePathSegments,
)


class Url(serializemixin.SerializeMixin, QtCore.QUrl):
    """Convenient interface for working with URLs."""

    def __init__(self, *args, **kwargs):
        match args:
            case (os.PathLike(), *rest):
                path = os.fspath(args[0])
                if pathlib.Path(path).exists():
                    path = self.fromLocalFile(path)
                super().__init__(path, *rest, **kwargs)
            case _:
                super().__init__(*args, **kwargs)

    @property
    def _toString(self) -> str:
        return self.toString()

    __match_args__ = ("_toString",)

    # def __str__(self):
    #     return self.absolutePath()

    def __repr__(self):
        return get_repr(self, self.toString())

    def __str__(self):
        return self.toString()

    def __vfspath__(self) -> str:
        return self.toString()

    def serialize_fields(self):
        return dict(path=self.toString())

    def serialize(self) -> dict[str, Any]:
        return self.serialize_fields()

    def to_string(self) -> str:
        return self.toString()

    def to_path(self) -> pathlib.Path:
        """Get pathlib object from the URL.

        Returns:
            Path
        """
        return pathlib.Path(str(self))

    def is_local_file(self) -> bool:
        return self.isLocalFile()

    @classmethod
    def from_user_input(cls, url: str, working_dir: str | None = None) -> Self:
        if working_dir is None:
            working_dir = ""
        return cls(cls.fromUserInput(url, working_dir))

    @classmethod
    def from_local_file(cls, path: datatypes.PathType) -> Self:
        url = cls.fromLocalFile(os.fspath(path))
        return cls(url)

    def _has_explicit_scheme(self) -> bool:
        """Check if a url has an explicit scheme given."""
        return bool(
            self.isValid()
            and self.scheme()
            and (self.host() or self.path())
            and not self.path().startswith(":")
        )

    def is_special_url(self) -> bool:
        """Return True if url is an about:... or other special URL."""
        return self.scheme() in ("about", "file") if self.isValid() else False


if __name__ == "__main__":
    url = Url()
    str(url)



from __future__ import annotations

import logging

from prettyqt import constants, core, gui, itemmodels
from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)


class SliceChangeIconSizeProxyModel(itemmodels.SliceIdentityProxyModel):
    """Proxy model which changes the icon size of the Decoration role.

    Supports QColors, QPixmaps and QIcons in DecorationRole.

    ### Example

    ```py
    model = MyModel()
    table = widgets.TreeView()
    table.set_model(model)
    table[0].proxify.change_icon_size(size=20)
    table.show()
    # or
    proxy = itemmodels.SliceFilterProxyModel(indexer=0, size=20)
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```

    === "Without proxy"

        ```py
        table = widgets.TreeView()
        source_model = widgets.FileSystemModel()
        ...
        table.set_model(source_model)
        # table.proxifier.change_icon_size(size=(30, 30))
        ```
        <figure markdown>
          ![Image title](../../images/slicechangeiconsizeproxymodel_before.png)
        </figure>

    === "With proxy"

        ```py
        table = widgets.TreeView()
        source_model = widgets.FileSystemModel()
        ...
        table.set_model(source_model)
        table.proxifier.change_icon_size(size=(30, 30))
        ```
        <figure markdown>
          ![Image title](../../images/slicechangeiconsizeproxymodel_after.png)
        </figure>

    """

    ID = "change_icon_size"
    ICON = "mdi.resize"

    def __init__(self, size: datatypes.SizeType, **kwargs):
        super().__init__(**kwargs)
        self._size = datatypes.to_size(size)
        self._cache = {}

    def data(  # noqa: PLR0911
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if role == constants.DECORATION_ROLE and self.indexer_contains(index):
            original = super().data(index, role)
            match original:
                case gui.QIcon():
                    hashed = original.cacheKey()
                    if hashed in self._cache:
                        return self._cache[hashed]
                    p = original.pixmap(self._size)
                    self._cache[hashed] = p
                    return p
                case gui.QColor():
                    hashed = original.name()
                    if hashed in self._cache:
                        return self._cache[hashed]
                    p = gui.QPixmap(self._size)
                    p.fill(original)
                    self._cache[hashed] = p
                    return p
                case gui.QPixmap():
                    hashed = original.cacheKey()
                    if hashed in self._cache:
                        return self._cache[hashed]
                    p = original.scaled(self._size)
                    self._cache[hashed] = p
                    return p
        return super().data(index, role)

    def set_icon_size(self, size: core.QSize):
        self._cache = {}
        self._size = size

    def get_icon_size(self) -> core.QSize:
        return self._size

    icon_size = core.Property(
        core.QSize,
        get_icon_size,
        set_icon_size,
        doc="New icon size",
    )


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    tree = widgets.TreeView()
    # tree.setUniformRowHeights(False)
    model = widgets.FileSystemModel()
    tree.set_model(model)
    tree.show()
    tree.set_sorting_enabled(True)
    tree.resize(650, 400)
    tree.set_icon("mdi.file-cabinet")
    tree.setWindowTitle("Example")
    index = model.setRootPath("E:/")
    # tree.setRootIndex(index)
    tree.h_header.resize_sections("stretch")
    # tree.proxifier.get_proxy("change_icon_size", size=(30, 30), indexer=None)

    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtSpatialAudio
from prettyqt.utils import bidict, datatypes


DistanceModelStr = Literal[
    "logarithmic",
    "linear",
    "manual_attenuation",
]

DISTANCE_MODEL: bidict[DistanceModelStr, QtSpatialAudio.QSpatialSound.DistanceModel] = (
    bidict(
        logarithmic=QtSpatialAudio.QSpatialSound.DistanceModel.Logarithmic,
        linear=QtSpatialAudio.QSpatialSound.DistanceModel.Linear,
        manual_attenuation=QtSpatialAudio.QSpatialSound.DistanceModel.ManualAttenuation,
    )
)

LoopsStr = Literal["infinite", "once"]

LOOPS: bidict[LoopsStr, QtSpatialAudio.QSpatialSound.Loops] = bidict(
    infinite=QtSpatialAudio.QSpatialSound.Loops.Infinite,
    once=QtSpatialAudio.QSpatialSound.Loops.Once,
)


class SpatialSound(core.ObjectMixin, QtSpatialAudio.QSpatialSound):
    """A sound object in 3D space."""

    def set_distance_model(
        self, model: DistanceModelStr | QtSpatialAudio.QSpatialSound.DistanceModel
    ):
        """Set the distance model.

        Args:
            model: distance model
        """
        self.setDistanceModel(DISTANCE_MODEL.get_enum_value(model))

    def get_distance_model(self) -> DistanceModelStr:
        """Return current distance model.

        Returns:
            distance model
        """
        return DISTANCE_MODEL.inverse[self.distanceModel()]

    def set_position(self, position: datatypes.Vector3DType):
        self.setPosition(datatypes.to_vector3d(position))



# credits to
# https://stackoverflow.com/questions/48425316/how-to-create-pyqt-properties-dynamically
from __future__ import annotations

import functools
from typing import ClassVar

from prettyqt.qt import QtCore


class PropertyMeta(type(QtCore.QObject)):
    """Lets a class succinctly define Qt properties."""

    def __new__(cls, name, bases, attrs):
        for key in list(attrs.keys()):
            attr = attrs[key]
            if not isinstance(attr, SyncedProperty):
                continue

            types = {list: "QVariantList", dict: "QVariantMap"}
            type_ = types.get(attr.type_, attr.type_)

            notifier = QtCore.Signal(type_)
            attrs[f"_{key}_changed"] = notifier
            attrs[key] = PropertyImpl(type_=type_, name=key, notify=notifier)

        return super().__new__(cls, name, bases, attrs)


class SyncedProperty:
    """Property definition.

    Instances of this class will be replaced with their full
    implementation by the PropertyMeta metaclass.
    """

    def __init__(self, type_):
        self.type_ = type_


class PropertyImpl(QtCore.Property):
    """Property implementation: gets, sets, and notifies of change."""

    def __init__(self, type_, name, notify):
        super().__init__(type_, self.getter, self.setter, notify=notify)
        self.name = name

    def getter(self, instance):
        return getattr(instance, f"_{self.name}")

    def setter(self, instance, value):
        signal = getattr(instance, f"_{self.name}_changed")

        if type(value) in {list, dict}:
            value = make_notified(value, signal)

        setattr(instance, f"_{self.name}", value)
        signal.emit(value)


class MakeNotified:
    """Adds notifying signals to lists and dictionaries.

    Creates the modified classes just once, on initialization.
    """

    change_methods: ClassVar[dict[type, list[str]]] = {
        list: [
            "__delitem__",
            "__iadd__",
            "__imul__",
            "__setitem__",
            "append",
            "extend",
            "insert",
            "pop",
            "remove",
            "reverse",
            "sort",
        ],
        dict: [
            "__delitem__",
            "__ior__",
            "__setitem__",
            "clear",
            "pop",
            "popitem",
            "setdefault",
            "update",
        ],
    }

    def __init__(self):
        self.notified_class = {
            type_: self.make_notified_class(type_) for type_ in [list, dict]
        }

    def __call__(self, seq, signal):
        """Returns a notifying version of the supplied list or dict."""
        notified_class = self.notified_class[type(seq)]
        notified_seq = notified_class(seq)
        notified_seq.signal = signal
        return notified_seq

    @classmethod
    def make_notified_class(cls, parent):
        notified_class = type(f"notified_{parent.__name__}", (parent,), {})
        for method_name in cls.change_methods[parent]:
            original = getattr(notified_class, method_name)
            notified_method = cls.make_notified_method(original, parent)
            setattr(notified_class, method_name, notified_method)
        return notified_class

    @staticmethod
    def make_notified_method(method, parent):
        @functools.wraps(method)
        def notified_method(self, *args, **kwargs):
            result = getattr(parent, method.__name__)(self, *args, **kwargs)
            self.signal.emit(self)
            return result

        return notified_method


make_notified = MakeNotified()


if __name__ == "__main__":

    class Demo(QtCore.QObject, metaclass=PropertyMeta):
        number = SyncedProperty(float)
        things = SyncedProperty(list)

        def __init__(self, parent=None):
            super().__init__(parent)
            self.number = 3.14

    demo1 = Demo()
    demo2 = Demo()
    demo1.number = 2.7
    demo1.things = ["spam", "spam", "baked beans", "spam"]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self

import anyenv

from prettyqt import core


if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from prettyqt.utils import datatypes


DB = core.MimeDatabase()


class MimeData(core.ObjectMixin, core.QMimeData):
    """Container for data that records information about its MIME type."""

    def __len__(self):
        return len(self.formats())

    # def __getitem__(self, index: str) -> str:
    #     return self.get_data(index)

    # def __setitem__(self, index: str, value: datatypes.ByteArrayType):
    #     value = datatypes.to_bytearray(value)
    #     self.setData(index, value)

    def __contains__(self, fmt: str):
        return self.hasFormat(fmt)

    def __delitem__(self, index: str):
        self.removeFormat(index)

    def set_data(self, mime_type: str, data: str):
        self.setData(mime_type, core.QByteArray(data.encode()))

    def set_json_data(self, mime_type: str, data: datatypes.JSONType):
        bytes_ = anyenv.dump_json(data).encode()
        self.setData(mime_type, core.QByteArray(bytes_))

    def get_data(self, mime_type: str) -> str:
        return bytes(self.data(mime_type)).decode()

    def get_json_data(self, mime_type: str) -> datatypes.JSONType:
        data = self.data(mime_type)
        return anyenv.load_json(bytes(data))

    def keys(self) -> list[str]:
        return self.formats()

    def values(self) -> Iterator[Any]:
        return (self.get_data(key) for key in self.formats())

    def set_data_for_extension(self, extension: str, string: str):
        if mimetype := DB.get_mime_types_for_filename(f".{extension}"):
            self.setData(mimetype[0].name(), string.encode())
        else:
            raise ValueError(extension)

    def set_path_data(self, paths: Iterable[datatypes.PathType]):
        urls = [core.Url.from_local_file(p) for p in paths]
        self.setUrls(urls)

    def set_urls(self, paths: Iterable[datatypes.PathType]):
        urls = [core.Url(p) for p in paths]
        self.setUrls(urls)

    def get_urls(self) -> list[core.Url]:
        return [core.Url(url) for url in self.urls()]

    @classmethod
    def for_file(
        cls,
        path: datatypes.PathType | core.QFileInfo,
        match_mode: core.mimedatabase.MatchModeStr = "default",
    ) -> MimeData:
        db = core.MimeDatabase()
        mime_type = db.get_mime_type_for_file(path, match_mode)
        return cls(mime_type)

    def to_dict(self) -> dict[str, bytes]:
        return {i: self.data(i).data() for i in self.formats()}

    @classmethod
    def clone(cls, other: core.QMimeData) -> Self:
        mime = cls()
        for fmt in other.formats():
            mime.setData(fmt, other.data(fmt))
        return mime


if __name__ == "__main__":
    from prettyqt import gui

    app = gui.app()
    c = app.get_clipboard()
    mime_data = c.mimeData()
    mime_data = MimeData.clone(mime_data)
    print(mime_data.to_dict())



from __future__ import annotations

from prettyqt import core, gui, widgets
from prettyqt.utils import get_repr


class FontChooserButton(widgets.Widget):
    value_changed = core.Signal(gui.QFont)

    def __init__(
        self,
        font: gui.QFont | str | None = None,
        object_name: str = "font_chooser_button",
        **kwargs,
    ):
        super().__init__(object_name=object_name, **kwargs)
        self._current_font = None
        layout = self.set_layout("horizontal", margin=0)
        self.lineedit = widgets.LineEdit(read_only=True)
        self.set_current_font(font)
        layout.add(self.lineedit)
        action = gui.Action(icon="mdi.format-font", triggered=self.choose_font)
        self.button = widgets.ToolButton()
        self.button.setDefaultAction(action)
        layout.add(self.button)

    def __repr__(self):
        return get_repr(self, self.current_font)

    @core.Slot()
    def choose_font(self):
        dlg = widgets.FontDialog()
        if self._current_font:
            dlg.setCurrentFont(self._current_font)

        if dlg.exec():
            self.set_current_font(dlg.current_font())
            self.value_changed.emit(gui.QFont(dlg.current_font()))

    def set_current_font(self, font: str | gui.QFont | None):
        match font:
            case str():
                self._current_font = gui.Font(font)
            case None:
                self._current_font = gui.Font()
            case _:
                self._current_font = font
        self.lineedit.setText(self._current_font.family())

    def set_value(self, value: str | gui.QFont):
        self.set_current_font(value)

    def get_value(self) -> gui.QFont:
        return gui.QFont(self._current_font)

    current_font = core.Property(
        gui.QFont,
        get_value,
        set_value,
        user=True,
        doc="Currently chosen font",
    )


if __name__ == "__main__":
    app = widgets.app()
    btn = FontChooserButton()
    btn.set_current_font(gui.Font.mono())
    btn.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtWebEngineCore
from prettyqt.utils import bidict


ROLE = bidict(
    url=QtWebEngineCore.QWebEngineHistoryModel.Roles.UrlRole,
    title=QtWebEngineCore.QWebEngineHistoryModel.Roles.TitleRole,
    offset=QtWebEngineCore.QWebEngineHistoryModel.Roles.OffsetRole,
    icon_url=QtWebEngineCore.QWebEngineHistoryModel.Roles.IconUrlRole,
)

RoleStr = Literal[
    "url",
    "title",
    "offset",
    "icon_url",
]

WebEngineHistoryModel = QtWebEngineCore.QWebEngineHistoryModel

# cant subclass.

# class WebEngineHistoryModel(
#     core.AbstractListModelMixin, QtWebEngineCore.QWebEngineHistoryModel
# ):
#     pass


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    model = WebEngineHistoryModel()
    widget = widgets.TableView()
    # widget.set_model(model)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import widgets


if TYPE_CHECKING:
    from prettyqt import core, gui


# https://stackoverflow.com/a/55252650/3620725


class NoFocusDelegate(widgets.StyledItemDelegate):
    """Delegate to remove dotted border on cell focus."""

    ID = "no_focus"

    def paint(
        self,
        painter: gui.QPainter,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        if option.state & widgets.Style.StateFlag.State_HasFocus:
            option.state = option.state ^ widgets.Style.StateFlag.State_HasFocus
        super().paint(painter, option, index)



from __future__ import annotations

from prettyqt import widgets
from prettyqt.utils import bidict


mod = widgets.QStyleOptionHeader

SECTION_POSITION = bidict(
    beginning=mod.SectionPosition.Beginning,
    middle=mod.SectionPosition.Middle,
    end=mod.SectionPosition.End,
    only_one_section=mod.SectionPosition.OnlyOneSection,
)

SELECTED_POSITION = bidict(
    not_adjacent=mod.SelectedPosition.NotAdjacent,
    next_is_selected=mod.SelectedPosition.NextIsSelected,
    previous_is_selected=mod.SelectedPosition.PreviousIsSelected,
    next_and_previous_are_selected=mod.SelectedPosition.NextAndPreviousAreSelected,
)

SORT_INDICATOR = bidict(
    none=mod.SortIndicator(0),  # type: ignore
    sort_up=mod.SortIndicator.SortUp,
    sort_down=mod.SortIndicator.SortDown,
)


class StyleOptionHeader(widgets.StyleOptionMixin, widgets.QStyleOptionHeader):
    pass



from __future__ import annotations

import re
from typing import ClassVar

from prettyqt import gui


class BaseScientificValidator(gui.Validator):
    re_pattern: re.Pattern
    group_map: dict[str, int]

    def get_group_dict(self, string: str):
        """Match the input string with the regex of this validator.

        The match groups will be put into a dict with strings  as keys describing
        the role of the specific group (i.e. mantissa, exponent, si-prefix etc.)

        Arguments:
            string: input string to be matched

        Returns:
            dictionary containing groups as items and strings as keys
        """
        match = self.re_pattern.search(string)
        if not match:
            return False
        groups = match.groups()
        return {k: groups[self.group_map[k]] for k in self.group_map}

    def fixup(self, text):
        return (
            match.groups()[0].strip() if (match := self.re_pattern.search(text)) else ""
        )


class ScientificIntegerValidator(BaseScientificValidator):
    """This is a validator for int values represented as strings in scientific notation.

    Using engineering notation only positive exponents are allowed
    (i.e. "1e9", "2E+8", "14e+3" etc.)
    Also supports non-fractional SI unit prefix like 'M', 'k' etc.
    """

    ID = "scientific_integer"

    re_pattern = re.compile(
        r"(([+-]?\d+)([eE]\+?\d+)?\s?([YZEPTGMk])?\s*)", flags=re.UNICODE
    )
    group_map: ClassVar = {"match": 0, "mantissa": 1, "exponent": 2, "si": 3}

    def validate(self, string, position):
        if not string.strip() or string.strip() in list("YZEPTGMk"):
            return self.State.Intermediate, string, position

        if not (group_dict := self.get_group_dict(string)):
            return self.State.Invalid, "", position
        if group_dict["match"] == string:
            return self.State.Acceptable, string, position

        position = min(position, len(string))
        if string[position - 1] in "eE-+":
            return self.State.Intermediate, string, position

        return self.State.Invalid, group_dict["match"], position


class ScientificFloatValidator(BaseScientificValidator):
    """This is a validator for float values represented as strings in scientific notation.

    (i.e. "1.35e-9", ".24E+8", "14e3" etc.)
    Also supports SI unit prefix like 'M', 'n' etc.
    """

    ID = "scientific_float"
    re_pattern = re.compile(
        r"(\s*([+-]?)(\d+\.\d+|\.\d+|\d+\.?)([eE][+-]?\d+)?\s?([YZEPTGMkmµunpfazy]?)\s*)",
        flags=re.UNICODE,
    )
    group_map: ClassVar = {"match": 0, "sign": 1, "mantissa": 2, "exponent": 3, "si": 4}

    def validate(self, string, position):  # noqa: PLR0911
        if (
            string.strip() in "+.-."
            or string.strip() in list("YZEPTGMkmµunpfazy")
            or re.match(r"[+-]?(in$|i$)", string, re.IGNORECASE)
        ):
            return self.State.Intermediate, string, position

        # Accept input of [+-]inf. Not case sensitive.
        if re.match(r"[+-]?\binf$", string, re.IGNORECASE):
            return self.State.Acceptable, string.lower(), position

        if not (group_dict := self.get_group_dict(string)):
            return (
                (self.State.Intermediate, string, position)
                if string[position - 1] in "eE-+." and "i" not in string.lower()
                else (self.State.Invalid, "", position)
            )
        if group_dict["match"] == string:
            return self.State.Acceptable, string, position
        if string.count(".") > 1:
            return self.State.Invalid, group_dict["match"], position
        position = min(position, len(string))
        if string[position - 1] in "eE-+" and "i" not in string.lower():
            return self.State.Intermediate, string, position
        return self.State.Invalid, group_dict["match"], position



from __future__ import annotations

import hashlib
import logging
import pathlib
from typing import TYPE_CHECKING, ClassVar, Literal

from prettyqt.qt import QtGui
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)

SYSTEM_FONT = bidict(
    general=QtGui.QFontDatabase.SystemFont.GeneralFont,
    fixed=QtGui.QFontDatabase.SystemFont.FixedFont,
    title=QtGui.QFontDatabase.SystemFont.TitleFont,
    smallest_readable=QtGui.QFontDatabase.SystemFont.SmallestReadableFont,
)

SystemFontStr = Literal["general", "fixed", "title", "smallest_readable"]

WRITING_SYSTEM = bidict(
    any=QtGui.QFontDatabase.WritingSystem.Any,
    latin=QtGui.QFontDatabase.WritingSystem.Latin,
    greek=QtGui.QFontDatabase.WritingSystem.Greek,
    cyrillic=QtGui.QFontDatabase.WritingSystem.Cyrillic,
    armenian=QtGui.QFontDatabase.WritingSystem.Armenian,
    hebrew=QtGui.QFontDatabase.WritingSystem.Hebrew,
    arabic=QtGui.QFontDatabase.WritingSystem.Arabic,
    syriac=QtGui.QFontDatabase.WritingSystem.Syriac,
    thaana=QtGui.QFontDatabase.WritingSystem.Thaana,
    devanagari=QtGui.QFontDatabase.WritingSystem.Devanagari,
    bengali=QtGui.QFontDatabase.WritingSystem.Bengali,
    gurmukhi=QtGui.QFontDatabase.WritingSystem.Gurmukhi,
    gujarati=QtGui.QFontDatabase.WritingSystem.Gujarati,
    oriya=QtGui.QFontDatabase.WritingSystem.Oriya,
    tamil=QtGui.QFontDatabase.WritingSystem.Tamil,
    telugu=QtGui.QFontDatabase.WritingSystem.Telugu,
    kannada=QtGui.QFontDatabase.WritingSystem.Kannada,
    malayalam=QtGui.QFontDatabase.WritingSystem.Malayalam,
    sinhala=QtGui.QFontDatabase.WritingSystem.Sinhala,
    thai=QtGui.QFontDatabase.WritingSystem.Thai,
    lao=QtGui.QFontDatabase.WritingSystem.Lao,
    tibetan=QtGui.QFontDatabase.WritingSystem.Tibetan,
    myanmar=QtGui.QFontDatabase.WritingSystem.Myanmar,
    georgian=QtGui.QFontDatabase.WritingSystem.Georgian,
    khmer=QtGui.QFontDatabase.WritingSystem.Khmer,
    simplified_chinese=QtGui.QFontDatabase.WritingSystem.SimplifiedChinese,
    traditional_chinese=QtGui.QFontDatabase.WritingSystem.TraditionalChinese,
    japanese=QtGui.QFontDatabase.WritingSystem.Japanese,
    korean=QtGui.QFontDatabase.WritingSystem.Korean,
    vietnamese=QtGui.QFontDatabase.WritingSystem.Vietnamese,
    symbol=QtGui.QFontDatabase.WritingSystem.Symbol,
    ogham=QtGui.QFontDatabase.WritingSystem.Ogham,
    runic=QtGui.QFontDatabase.WritingSystem.Runic,
    nko=QtGui.QFontDatabase.WritingSystem.Nko,
)

WritingSystemStr = Literal[
    "any",
    "latin",
    "greek",
    "cyrillic",
    "armenian",
    "hebrew",
    "arabic",
    "syriac",
    "thaana",
    "devanagari",
    "bengali",
    "gurmukhi",
    "gujarati",
    "oriya",
    "tamil",
    "telugu",
    "kannada",
    "malayalam",
    "sinhala",
    "thai",
    "lao",
    "tibetan",
    "myanmar",
    "georgian",
    "khmer",
    "simplified_chinese",
    "traditional_chinese",
    "japanese",
    "korean",
    "vietnamese",
    "symbol",
    "ogham",
    "runic",
    "nko",
]


class FontDatabase(QtGui.QFontDatabase):
    """Information about the fonts available in the underlying window system."""

    font_paths: ClassVar[dict[str, int]] = {}

    @classmethod
    def add_fonts_from_folder(cls, path: datatypes.PathType):
        path = pathlib.Path(path)
        for p in path.iterdir():
            if p.suffix.lower() in [".ttf", ".otf"]:
                logger.debug("adding font %r to database.", p)
                cls.addApplicationFont(str(p))

    @classmethod
    def add_font(cls, path: datatypes.PathType, ttf_hash: str | None = None) -> int:
        path = pathlib.Path(path)
        font_id = cls.addApplicationFont(str(path))
        if not cls.applicationFontFamilies(font_id):
            msg = (
                f"Font {path!r} appears to be empty. "
                "If you are on Windows 10, please read "
                "https://support.microsoft.com/"
                "en-us/kb/3053676"
            )
            raise RuntimeError(msg)
        if ttf_hash is not None:
            content = path.read_bytes()
            if hashlib.md5(content).hexdigest() != ttf_hash:
                msg = f"Font is corrupt at: {path!r}"
                raise OSError(msg)
        cls.font_paths[str(path)] = font_id
        return font_id

    @classmethod
    def remove_font(cls, font: datatypes.PathType | int):
        font_id = font if isinstance(font, int) else cls.font_paths[str(font)]
        cls.removeApplicationFont(font_id)

    @classmethod
    def get_system_font(cls, font_type: SystemFontStr | QtGui.QFontDatabase.SystemFont):
        return cls.systemFont(SYSTEM_FONT[font_type])


if __name__ == "__main__":
    from prettyqt import gui

    app = gui.app()
    db = FontDatabase()



from __future__ import annotations

from prettyqt import gui
from prettyqt.qt import QtWidgets


class PlainTextDocumentLayout(
    gui.AbstractTextDocumentLayoutMixin, QtWidgets.QPlainTextDocumentLayout
):
    """Implements a plain text layout for QTextDocument."""


if __name__ == "__main__":
    doc = gui.TextDocument()
    layout = PlainTextDocumentLayout(doc)



from __future__ import annotations

from prettyqt import core, qthelp


class HelpContentModel(core.AbstractItemModelMixin):
    """Model that supplies content to views."""

    def __init__(self, item: qthelp.QHelpContentModel):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_content_item_at(self, index) -> qthelp.HelpContentItem:
        return qthelp.HelpContentItem(self.contentItemAt(index))



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtGui


class Vector4D(QtGui.QVector4D):
    def __bool__(self):
        return not self.isNull()

    def __abs__(self) -> float:
        return self.length()

    @property
    def _x(self) -> float:
        return self.x()

    @property
    def _y(self) -> float:
        return self.y()

    @property
    def _z(self) -> float:
        return self.z()

    @property
    def _w(self) -> float:
        return self.w()

    __match_args__ = ("_x", "_y", "z", "w")

    def __reduce__(self):
        return type(self), (self.x(), self.y(), self.z(), self.w())

    def to_point(self) -> core.Point:
        return core.Point(self.toPoint())

    def to_pointf(self) -> core.PointF:
        return core.PointF(self.toPointF())


if __name__ == "__main__":
    vector = Vector4D(0, 0, 0, 1)



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import bidict


TickPositionStr = Literal["none", "both_sides", "above", "below"]

TICK_POSITION: bidict[TickPositionStr, widgets.QSlider.TickPosition] = bidict(
    none=widgets.QSlider.TickPosition.NoTicks,
    both_sides=widgets.QSlider.TickPosition.TicksBothSides,
    above=widgets.QSlider.TickPosition.TicksAbove,
    below=widgets.QSlider.TickPosition.TicksBelow,
)

TickPositionAllStr = Literal["none", "both_sides", "above", "below", "left", "right"]


class HollowHandleStyle(widgets.ProxyStyle):
    def __init__(self, config: dict | None = None):
        super().__init__()
        self.config = {
            "groove.height": 3,
            "sub-page.color": gui.Color(255, 255, 255),
            "add-page.color": gui.Color(255, 255, 255, 64),
            "handle.color": gui.Color(255, 255, 255),
            "handle.ring-width": 4,
            "handle.hollow-radius": 6,
            "handle.margin": 4,
        }
        config = config or {}
        self.config |= config

        # get handle size
        w = (
            self.config["handle.margin"]
            + self.config["handle.ring-width"]
            + self.config["handle.hollow-radius"]
        )
        self.config["handle.size"] = core.QSize(2 * w, 2 * w)

    def subControlRect(
        self,
        cc: widgets.QStyle.ComplexControl,
        opt: widgets.QStyleOptionSlider,
        sc: widgets.QStyle.SubControl,
        widget: widgets.QWidget,
    ):
        """Get the rectangular area occupied by the sub control."""
        if (
            cc != self.ComplexControl.CC_Slider
            or opt.orientation != constants.Orientation.Horizontal
            or sc == self.SubControl.SC_SliderTickmarks
        ):
            return super().subControlRect(cc, opt, sc, widget)

        rect = opt.rect

        if sc == self.SubControl.SC_SliderGroove:
            h = self.config["groove.height"]
            groove_rect = core.QRectF(0, (rect.height() - h) // 2, rect.width(), h)
            return groove_rect.toRect()

        if sc == self.SubControl.SC_SliderHandle:
            size = self.config["handle.size"]
            x = self.sliderPositionFromValue(
                opt.minimum, opt.maximum, opt.sliderPosition, rect.width()
            )

            # solve the situation that the handle runs out of slider
            x *= (rect.width() - size.width()) / rect.width()
            slider_rect = core.QRectF(x, 0, size.width(), size.height())
            return slider_rect.toRect()
        return None

    def drawComplexControl(
        self,
        cc: widgets.QStyle.ComplexControl,
        opt: widgets.QStyleOptionSlider,
        painter: gui.QPainter,
        widget: widgets.QWidget,
    ):
        """Draw sub control."""
        if (
            cc != self.ComplexControl.CC_Slider
            or opt.orientation != constants.Orientation.Horizontal
        ):
            return super().drawComplexControl(cc, opt, painter, widget)

        groove_rect = self.subControlRect(
            cc, opt, self.SubControl.SC_SliderGroove, widget
        )
        handle_rect = self.subControlRect(
            cc, opt, self.SubControl.SC_SliderHandle, widget
        )
        painter.setRenderHints(gui.Painter.RenderHint.Antialiasing)
        painter.setPen(constants.PenStyle.NoPen)

        # paint groove
        painter.save()
        painter.translate(groove_rect.topLeft())

        # paint the crossed part
        w = handle_rect.x() - groove_rect.x()
        h = self.config["groove.height"]
        painter.setBrush(self.config["sub-page.color"])
        painter.drawRect(0, 0, w, h)

        # paint the uncrossed part
        x = w + self.config["handle.size"].width()
        painter.setBrush(self.config["add-page.color"])
        painter.drawRect(x, 0, groove_rect.width() - w, h)
        painter.restore()

        # paint handle
        ring_width = self.config["handle.ring-width"]
        hollow_radius = self.config["handle.hollow-radius"]
        radius = ring_width + hollow_radius

        path = gui.PainterPath()
        path.moveTo(0, 0)
        center = handle_rect.center().toPointF() + core.QPointF(1, 1)
        path.addEllipse(center, radius, radius)
        path.addEllipse(center, hollow_radius, hollow_radius)

        handle_color = self.config["handle.color"]  # type:QColor
        handle_color.setAlpha(
            255 if opt.activeSubControls != self.SubControl.SC_SliderHandle else 153
        )
        painter.setBrush(handle_color)
        painter.drawPath(path)

        # press handle
        if widget.isSliderDown():
            handle_color.setAlpha(255)
            painter.setBrush(handle_color)
            painter.drawEllipse(handle_rect)
            return None
        return None


class Slider(widgets.AbstractSliderMixin, widgets.QSlider):
    """Vertical or horizontal slider."""

    value_changed = core.Signal(int)
    clicked = core.Signal(int)

    def __init__(self, *args, **kwargs):
        match args:
            case (str(), *rest):
                super().__init__(constants.ORIENTATION[args[0]], *rest, **kwargs)
            case _:
                super().__init__(*args, **kwargs)
        self.valueChanged.connect(self.on_value_change)
        # style = HollowHandleStyle(
        #     {
        #         "groove.height": 4,
        #         "sub-page.color": gui.QColor(72, 210, 242),
        #         "add-page.color": gui.QColor(255, 255, 255, 50),
        #         "handle.color": gui.QColor(72, 210, 242),
        #         "handle.ring-width": 2,
        #         "handle.hollow-radius": 10,
        #         "handle.margin": 0,
        #     }
        # )
        # self.setStyle(style)

    def mousePressEvent(self, e: gui.QMouseEvent):
        self.clicked.emit(self.value())
        if self.orientation() == constants.Orientation.Horizontal:
            value = e.pos().x() / self.width() * self.maximum()
        else:
            value = (self.height() - e.pos().y()) / self.height() * self.maximum()

        self.setValue(int(value))
        super().mousePressEvent(e)

    def set_tick_position(
        self, position: TickPositionAllStr | widgets.QSlider.TickPosition
    ):
        """Set the tick position for the slider.

        For vertical orientation, "above" equals to "left" and "below" to "right".

        Args:
            position: position for the ticks
        """
        if position == "left":
            position = "above"
        elif position == "right":
            position = "below"
        self.setTickPosition(TICK_POSITION.get_enum_value(position))

    def get_tick_position(self) -> TickPositionStr:
        """Return tick position.

        Returns:
            tick position
        """
        return TICK_POSITION.inverse[self.tickPosition()]
        # if self.is_vertical():
        #     if val == "above":
        #         return "left"
        #     elif val == "below":
        #         return "right"


if __name__ == "__main__":
    app = widgets.app()
    slider = Slider()
    slider.setRange(0, 100)
    slider.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import bluetooth, core
from prettyqt.utils import bidict


Error = bluetooth.QBluetoothServer.Error

ErrorStr = Literal[
    "none",
    "unknown",
    "powered_off",
    "input_output",
    "service_already_registered",
    "unsupported_protocol",
    "missing_permissions",
]

ERROR: bidict[ErrorStr, Error] = bidict(
    none=Error.NoError,
    unknown=Error.UnknownError,
    powered_off=Error.PoweredOffError,
    input_output=Error.InputOutputError,
    service_already_registered=Error.ServiceAlreadyRegisteredError,
    unsupported_protocol=Error.UnsupportedProtocolError,
    missing_permissions=Error.MissingPermissionsError,
)


class BluetoothServer(core.ObjectMixin, bluetooth.QBluetoothServer):
    """Uses the RFCOMM or L2cap protocol to communicate with a Bluetooth device."""

    def __init__(
        self,
        protocol: bluetooth.bluetoothserviceinfo.ProtocolStr
        | bluetooth.QBluetoothServiceInfo.Protocol,
        parent: core.QObject | None = None,
    ):
        protocol = bluetooth.bluetoothserviceinfo.PROTOCOL.get_enum_value(protocol)
        super().__init__(protocol, parent)

    def get_error(self) -> ErrorStr:
        """Get error code."""
        return ERROR.inverse[self.error()]

    def get_server_type(self) -> bluetooth.bluetoothserviceinfo.ProtocolStr:
        """Get server type."""
        return bluetooth.bluetoothserviceinfo.PROTOCOL.inverse[self.serverType()]

    def get_server_address(self) -> bluetooth.BluetoothAddress:
        """Return bluetooth server address."""
        return bluetooth.BluetoothAddress(self.serverAddress())


if __name__ == "__main__":
    from prettyqt import core

    app = core.app()
    server = BluetoothServer("l2_cap")
    with app.debug_mode():
        app.exec()
